from dataclasses import asdict
from typing import Iterable, List, Optional, Union

from platypush.plugins import RunnablePlugin, action

from ._manager import AudioManager
from ._model import DeviceType, StreamType


class SoundPlugin(RunnablePlugin):
    """
    Plugin to interact with a sound device.

    Among the other features, enabling this plugin allows you to stream audio
    over HTTP from the device where the application is running, if
    :class:`platypush.backend.http.HttpBackend` is enabled.

    Simply open
    ``http://<host>:<backend-port>/sound/stream.[mp3|wav|acc|ogg]?token=<your-token>``
    to access live recording from the audio device.

    It can also be used as a general-purpose audio player and synthesizer,
    supporting both local and remote audio resources, as well as a MIDI-like
    interface through the :meth:`.play` command.
    """

    _DEFAULT_BLOCKSIZE = 1024
    _DEFAULT_QUEUE_SIZE = 10

    def __init__(
        self,
        input_device: Optional[DeviceType] = None,
        output_device: Optional[DeviceType] = None,
        input_blocksize: int = _DEFAULT_BLOCKSIZE,
        output_blocksize: int = _DEFAULT_BLOCKSIZE,
        queue_size: Optional[int] = _DEFAULT_QUEUE_SIZE,
        ffmpeg_bin: str = 'ffmpeg',
        **kwargs,
    ):
        """
        :param input_device: Index or name of the default input device. Use
            :meth:`.query_devices` to get the available devices. Default: system
            default
        :param output_device: Index or name of the default output device.
            Use :meth:`.query_devices` to get the available devices. Default:
            system default
        :param input_blocksize: Blocksize to be applied to the input device.
            Try to increase this value if you get input overflow errors while
            recording. Default: 1024
        :param output_blocksize: Blocksize to be applied to the output device.
            Try to increase this value if you get output underflow errors while
            playing. Default: 1024
        :param queue_size: When running in synth mode, this is the maximum
            number of generated audio frames that will be queued before the
            audio device processes them (default: 100).
        :param ffmpeg_bin: Path of the ``ffmpeg`` binary (default: search for
            the ``ffmpeg`` in the ``PATH``).
        """

        super().__init__(**kwargs)

        self.input_device = input_device
        self.output_device = output_device
        self.input_blocksize = input_blocksize
        self.output_blocksize = output_blocksize
        self.ffmpeg_bin = ffmpeg_bin
        self._manager = AudioManager(
            input_blocksize=self.input_blocksize,
            output_blocksize=self.output_blocksize,
            should_stop=self._should_stop,
            input_device=input_device,
            output_device=output_device,
            queue_size=queue_size,
        )

    @action
    def play(
        self,
        resource: Optional[str] = None,
        file: Optional[str] = None,
        sound: Optional[Union[dict, Iterable[dict]]] = None,
        device: Optional[DeviceType] = None,
        duration: Optional[float] = None,
        blocksize: Optional[int] = None,
        sample_rate: Optional[int] = None,
        channels: int = 2,
        volume: float = 100,
        dtype: Optional[str] = None,
        format: Optional[str] = None,  # pylint: disable=redefined-builtin
        stream_name: Optional[str] = None,
        stream_index: Optional[int] = None,
        join: bool = False,
    ):
        """
        Plays an audio file/URL (any audio format supported by ffmpeg works) or
        a synthetic sound.

        :param resource: Audio resource to be played. It can be a local file or
            a URL.
        :param file: **Deprecated**. Use ``resource`` instead.
        :param sound: Sound to play. Specify this if you want to play
            synthetic sounds. You can also create polyphonic sounds by just
            calling play multiple times. Frequencies can be specified either by
            ``midi_note`` - either as strings (e.g. ``A4``) or integers (e.g.
            ``69``) - or by ``frequency`` (e.g. ``440`` for 440 Hz). You can
            also specify a list of sounds here if you want to apply multiple
            harmonics on a base sound.
            Some examples:

                .. code-block:: python

                    {
                        "frequency": 440,  # 440 Hz
                        "volume":    100,  # Maximum volume
                        "duration":  1.0   # 1 second or until stop_playback
                    }

                .. code-block:: python

                    [
                        {
                            "midi_note": "A4", # A @ 440 Hz
                            "volume":    100,  # Maximum volume
                            "duration":  1.0   # 1 second or until stop_playback
                        },
                        {
                            "midi_note": "E5", # Play the harmonic one fifth up
                            "volume":    25,   # 1/4 of the volume
                            "duration":  1.0   # 1 second or until stop_playback
                            "phase":     3.14  # ~180 degrees phase
                            # Make it a triangular wave (default: sin).
                            # Supported types: "sin", "triang", "square",
                            # "sawtooth"
                            "shape:      "triang"
                        }
                    ]

                .. code-block:: python

                    [
                        {
                            "midi_note": "C4", # C4 MIDI note
                            "duration":  0.5   # 0.5 seconds or until stop_playback
                        },
                        {
                            "midi_note": "G5", # G5 MIDI note
                            "duration":  0.5,  # 0.5 seconds or until stop_playback
                            "delay":     0.5   # Start this note 0.5 seconds
                                               # after playback has started
                        }
                    ]

        :param device: Output device (default: default configured device or
            system default audio output if not configured)
        :param duration: Playback duration, in seconds. Default: None - play
            until the end of the audio source or until :meth:`.stop_playback`.
        :param blocksize: Audio block size (default: configured
            `output_blocksize` or 2048)
        :param sample_rate: Audio sample rate. Default: audio file sample rate
            if in file mode, 44100 Hz if in synth mode
        :param channels: Number of audio channels. Default: number of channels
            in the audio file in file mode, 1 if in synth mode
        :param volume: Playback volume, between 0 and 100. Default: 100.
        :param dtype: Data type for the audio samples, if playing raw PCM audio
            frames. Supported types: 'float64', 'float32', 'int32', 'int16',
            'int8', 'uint8'.
        :param format: Output audio format, if you want to convert the audio to
            another format before playing it. The list of available formats can
            be retrieved through the ``ffmpeg -formats`` command. Default: None
        :param stream_index: If specified, play to an already active stream
            index (you can get them through :meth:`.query_streams`). Default:
            creates a new audio stream through PortAudio.
        :param stream_name: Name of the stream to play to. If set, the sound
            will be played to the specified stream name, or a stream with that
            name will be created. If not set, and ``stream_index`` is not set
            either, then a new stream will be created on the next available
            index and named ``platypush-stream-<index>``.
        :param join: If True, then the method will block until the playback is
            completed. Default: False.
        """

        dev = self._manager.get_device(device=device, type=StreamType.OUTPUT)
        blocksize = blocksize or self.output_blocksize

        if file:
            self.logger.warning(
                'file is deprecated, use resource instead',
            )
            if not resource:
                resource = file

        if not (resource or sound):
            raise RuntimeError(
                'Please specify either a file to play or a list of sound objects'
            )

        self.logger.info(
            'Starting playback of %s to sound device [%s] on stream [%s]',
            resource or sound,
            dev.index,
            stream_index,
        )

        player_kwargs = {}
        if dtype:
            player_kwargs['dtype'] = dtype
        if format:
            player_kwargs['format'] = format

        player = self._manager.create_player(
            device=dev.index,
            infile=resource,
            sound=sound,
            duration=duration,
            blocksize=blocksize,
            sample_rate=sample_rate,
            channels=channels,
            volume=volume,
            stream_name=stream_name,
            **player_kwargs,
        )

        player.start()
        if join:
            player.join()

    @action
    def stream_recording(self, *args, **kwargs):
        """
        Deprecated alias for :meth:`.record`.
        """
        self.logger.warning(
            'sound.stream_recording is deprecated, use sound.record instead',
        )

        return self.record(*args, **kwargs)

    @action
    def record(
        self,
        device: Optional[DeviceType] = None,
        output_device: Optional[DeviceType] = None,
        fifo: Optional[str] = None,
        outfile: Optional[str] = None,
        duration: Optional[float] = None,
        sample_rate: Optional[int] = None,
        dtype: str = 'int16',
        blocksize: Optional[int] = None,
        latency: Union[float, str] = 'high',
        channels: int = 1,
        volume: float = 100,
        redis_queue: Optional[str] = None,
        format: str = 'wav',  # pylint: disable=redefined-builtin
        stream: bool = True,
        stream_name: Optional[str] = None,
        play_audio: bool = False,
    ):
        """
        Return audio data from an audio source

        :param device: Input device (default: default configured device or
            system default audio input if not configured)
        :param output_device: Audio output device if ``play_audio=True`` (audio
            pass-through) (default: default configured device or system default
            audio output if not configured)
        :param fifo: Path of a FIFO that will be used to exchange audio frames
            with other consumers.
        :param outfile: If specified, the audio data will be persisted on the
            specified audio file too.
        :param duration: Recording duration in seconds (default: record until
            stop event)
        :param sample_rate: Recording sample rate (default: device default rate)
        :param dtype: Data type for the audio samples. Supported types:
            'float64', 'float32', 'int32', 'int16', 'int8', 'uint8'. Default:
            int16.
        :param blocksize: Audio block size (default: configured
            `input_blocksize` or 2048)
        :param play_audio: If True, then the recorded audio will be played in
            real-time on the selected ``output_device`` (default: False).
        :param latency: Device latency in seconds (default: the device's default high latency)
        :param channels: Number of channels (default: 1)
        :param volume: Recording volume, between 0 and 100. Default: 100.
        :param redis_queue: If set, the audio chunks will also be published to
            this Redis channel, so other consumers can process them downstream.
        :param format: Audio format. Supported: wav, mp3, ogg, aac, flac.
            Default: wav.
        :param stream: If True (default), then the audio will be streamed to an
            HTTP endpoint too (default: ``/sound/stream<.format>``).
        :param stream_name: Custom name for the output stream.
        """

        dev = self._manager.get_device(device=device, type=StreamType.INPUT)
        self._manager.create_recorder(
            dev.index,
            output_device=output_device,
            fifo=fifo,
            outfile=outfile,
            duration=duration,
            sample_rate=sample_rate,
            dtype=dtype,
            blocksize=blocksize,
            latency=latency,
            channels=channels,
            volume=volume,
            redis_queue=redis_queue,
            format=format,
            stream=stream,
            stream_name=stream_name,
            play_audio=play_audio,
        ).start()

    @action
    def recordplay(self, *args, **kwargs):
        """
        Deprecated alias for :meth:`.record`.
        """
        self.logger.warning(
            'sound.recordplay is deprecated, use sound.record with `play_audio=True` instead',
        )

        kwargs['play_audio'] = True
        return self.record(*args, **kwargs)

    @action
    def status(self) -> List[dict]:
        """
        :return: The current status of the audio devices and streams.

        Example:

            .. code-block:: json

                [
                  {
                    "streams": [
                      {
                        "device": 3,
                        "direction": "output",
                        "outfile": "/dev/null",
                        "infile": "/mnt/hd/media/music/audio.mp3",
                        "ffmpeg_bin": "ffmpeg",
                        "channels": 2,
                        "sample_rate": 44100,
                        "dtype": "int16",
                        "streaming": false,
                        "duration": null,
                        "blocksize": 1024,
                        "latency": "high",
                        "redis_queue": "platypush-stream-AudioResourcePlayer-3",
                        "audio_pass_through": false,
                        "state": "PAUSED",
                        "started_time": "2023-06-19T11:57:05.882329",
                        "stream_index": 1,
                        "stream_name": "platypush:audio:output:1"
                      }
                    ],
                    "index": 3,
                    "name": "default",
                    "hostapi": 0,
                    "max_input_channels": 32,
                    "max_output_channels": 32,
                    "default_samplerate": 44100,
                    "default_low_input_latency": 0.008707482993197279,
                    "default_low_output_latency": 0.008707482993197279,
                    "default_high_input_latency": 0.034829931972789115,
                    "default_high_output_latency": 0.034829931972789115
                  }
                ]

        """
        devices = self._manager.get_devices()
        streams = self._manager.get_streams()
        ret = {dev.index: {'streams': [], **asdict(dev)} for dev in devices}
        for stream in streams:
            if stream.device is None:
                continue

            dev_index = int(
                stream.device
                if isinstance(stream.device, (int, str))
                else stream.device[0]
            )

            ret[dev_index]['streams'].append(stream.asdict())

        return list(ret.values())

    @action
    def query_streams(self):
        """
        Deprecated alias for :meth:`.status`.
        """

        self.logger.warning(
            'sound.query_streams is deprecated, use sound.status instead',
        )

        return self.status()

    @action
    def stop_playback(
        self,
        device: Optional[DeviceType] = None,
        streams: Optional[Iterable[Union[int, str]]] = None,
    ):
        """
        :param device: Only stop the streams on the specified device, by name or index (default: all).
        :param streams: Streams to stop by index or name (default: all).
        """
        self._manager.stop_audio(device=device, streams=streams, type=StreamType.OUTPUT)

    @action
    def pause_playback(
        self,
        device: Optional[DeviceType] = None,
        streams: Optional[Iterable[Union[int, str]]] = None,
    ):
        """
        :param device: Only stop the streams on the specified device, by name or index (default: all).
        :param streams: Streams to stop by index or name (default: all).
        """
        self._manager.pause_audio(
            device=device, streams=streams, type=StreamType.OUTPUT
        )

    @action
    def stop_recording(
        self, device: Optional[DeviceType] = None, timeout: Optional[float] = 2
    ):
        """
        Stop the current recording process on the selected device (default:
        default input device), if it is running.
        """
        self._manager.stop_audio(device, StreamType.INPUT, timeout=timeout)

    @action
    def pause_recording(self, device: Optional[DeviceType] = None):
        """
        Toggle the recording pause state on the selected device (default:
        default input device), if it is running.

        If paused, the recording will be resumed. If running, it will be
        paused. Otherwise, no action will be taken.
        """
        self._manager.pause_audio(device, StreamType.INPUT)

    @action
    def set_volume(
        self,
        volume: float,
        device: Optional[DeviceType] = None,
        streams: Optional[Iterable[Union[int, str]]] = None,
    ):
        """
        Set the audio input/output volume.

        :param volume: New volume, between 0 and 100.
        :param device: Set the volume only on the specified device (default:
            all).
        :param streams: Set the volume only on the specified list of stream
            indices/names (default: all).
        """
        self._manager.set_volume(volume=volume, device=device, streams=streams)

    def main(self):
        try:
            self.wait_stop()
        finally:
            self._manager.stop_audio()


# vim:sw=4:ts=4:et:



from typing import Optional

from sqlalchemy.orm import Session, exc

from platypush.entities import Entity


def get_parent(session: Session, entity: Entity) -> Optional[Entity]:
    """
    Gets the parent of an entity, and it fetches if it's not available in
    the current session.
    """
    try:
        return entity.parent
    except exc.DetachedInstanceError:
        # Dirty fix for `Parent instance <...> is not bound to a Session;
        # lazy load operation of attribute 'parent' cannot proceed`
        return session.query(Entity).get(entity.parent_id) if entity.parent_id else None



from platypush.message.event import Event


class AdafruitConnectedEvent(Event):
    """
    Event triggered when the backend connects to the Adafruit message queue.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class AdafruitDisconnectedEvent(Event):
    """
    Event triggered when the backend disconnects from the Adafruit message
    queue.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class AdafruitFeedUpdateEvent(Event):
    """
    Event triggered when a message is received on a subscribed Adafruit feed.
    """

    def __init__(self, feed, data, *args, **kwargs):
        super().__init__(*args, feed=feed, data=data, **kwargs)


# vim:sw=4:ts=4:et:



from abc import ABC
from datetime import datetime
from typing import Optional

from platypush.message.event import Event


class GPSEvent(Event, ABC):
    """
    Generic class for GPS events.
    """


class GPSDeviceEvent(GPSEvent):
    """
    Event triggered when a new GPS device is connected or reconfigured.
    """

    def __init__(
        self,
        path: str,
        *args,
        activated: Optional[datetime] = None,
        native: bool = False,
        baudrate: Optional[int] = None,
        parity: Optional[str] = None,
        stopbits: Optional[int] = None,
        cycle: Optional[float] = None,
        driver: Optional[str] = None,
        subtype: Optional[str] = None,
        **kwargs
    ):
        """
        :param path: Device path.
        :param activated: Device activation timestamp.
        :param native: Device native status.
        :param baudrate: Device baudrate.
        :param parity: Device parity.
        :param stopbits: Device stopbits.
        :param cycle: Device cycle.
        :param driver: Device driver.
        :param subtype: Device subtype.
        """
        super().__init__(
            *args,
            path=path,
            activated=activated,
            native=native,
            baudrate=baudrate,
            parity=parity,
            stopbits=stopbits,
            cycle=cycle,
            driver=driver,
            subtype=subtype,
            **kwargs
        )


class GPSLocationUpdateEvent(GPSEvent):
    """
    Event triggered upon GPS status update.
    """

    def __init__(
        self,
        *args,
        device=None,
        latitude=None,
        longitude=None,
        altitude=None,
        mode=None,
        **kwargs
    ):
        super().__init__(
            *args,
            device=device,
            latitude=latitude,
            longitude=longitude,
            altitude=altitude,
            mode=mode,
            **kwargs
        )


class GPSEnabledEvent(GPSEvent):
    """
    Event triggered when the GPS polling is enabled.
    """


class GPSDisabledEvent(GPSEvent):
    """
    Event triggered when the GPS polling is disabled.
    """


# vim:sw=4:ts=4:et:



from typing import Any

from platypush.message.event import Event


class WebsocketMessageEvent(Event):
    """
    Event triggered when a message is receive on a subscribed websocket URL.
    """

    def __init__(self, *args, url: str, message: Any, **kwargs):
        """
        :param url: Websocket URL.
        :param message: The received message.
        """
        super().__init__(*args, url=url, message=message, **kwargs)



"""
Platypush

.. moduleauthor:: Fabio Manganiello <fabio@manganiello.tech>
.. license: MIT
"""

from .app import Application, app
from .config import Config
from .context import Variable, get_backend, get_bus, get_plugin
from .cron import cron
from .event.hook import hook
from .message.event import Event
from .message.request import Request
from .message.response import Response
from .procedure import procedure
from .runner import main
from .utils import run

# Alias for platypush.event.hook.hook,
# see https://git.platypush.tech/platypush/platypush/issues/399
when = hook


__author__ = 'Fabio Manganiello <fabio@manganiello.tech>'
__version__ = '1.1.2'
__all__ = [
    'Application',
    'Variable',
    'Config',
    'Event',
    'Request',
    'Response',
    'app',
    'cron',
    'get_backend',
    'get_bus',
    'get_plugin',
    'hook',
    'main',
    'procedure',
    'run',
    'when',
]


# vim:sw=4:ts=4:et:



import base64
import io
import os
import threading
import time
from typing import Optional, List

import qrcode
from pyzbar import pyzbar
from PIL import Image

from platypush import Config
from platypush.context import get_bus
from platypush.message.event.qrcode import QrcodeScannedEvent
from platypush.plugins import Plugin, action
from platypush.plugins.camera import CameraPlugin
from platypush.schemas.qrcode import (
    QrcodeDecodedSchema,
    QrcodeDecodedResultSchema,
    QrcodeGeneratedSchema,
)
from platypush.utils import get_plugin_class_by_name


class QrcodePlugin(Plugin):
    """
    Plugin to generate and scan QR and bar codes.
    """

    def __init__(self, camera_plugin: Optional[str] = None, **kwargs):
        """
        :param camera_plugin: Name of the plugin that will be used as a camera
            to capture images (e.g. ``camera.cv`` or ``camera.pi``). This is
            required if you want to use the ``start_scanning`` action to scan
            QR codes from a camera.
        """
        super().__init__(**kwargs)
        self.camera_plugin = camera_plugin
        self._capturing = threading.Event()

    def _get_camera(
        self, camera_plugin: Optional[str] = None, **config
    ) -> CameraPlugin:
        camera_plugin = camera_plugin or self.camera_plugin
        assert camera_plugin, 'No camera plugin specified'
        if not config:
            config = Config.get(camera_plugin) or {}
        config['stream_raw_frames'] = True

        cls = get_plugin_class_by_name(camera_plugin)
        assert cls and issubclass(
            cls, CameraPlugin
        ), f'{camera_plugin} is not a valid camera plugin'
        return cls(**config)

    @action
    def generate(
        self,
        content: str,
        output_file: Optional[str] = None,
        show: bool = False,
        format: str = 'png',
    ) -> dict:
        """
        Generate a QR code.

        If you configured the :class:`platypush.backend.http.HttpBackend` then
        you can also generate codes directly from the browser through
        ``http://<host>:<port>/qrcode?content=...``.

        :param content: Text, URL or content of the QR code.
        :param output_file: If set then the QR code will be exported in the
            specified image file. Otherwise, a base64-encoded representation of
            its binary content will be returned in the response as ``data``.
        :param show: If True, and if the device where the application runs has
            an active display, then the generated QR code will be shown on
            display.
        :param format: Output image format (default: ``png``).
        :return: .. schema:: qrcode.QrcodeGeneratedSchema
        """
        qr = qrcode.make(content)
        img = qr.get_image()
        ret = {
            'content': content,
            'format': format,
        }

        if show:
            img.show()

        if output_file:
            output_file = os.path.abspath(os.path.expanduser(output_file))
            img.save(output_file, format=format)
            ret['image_file'] = output_file
        else:
            f = io.BytesIO()
            img.save(f, format=format)
            ret['data'] = base64.encodebytes(f.getvalue()).decode()

        return dict(QrcodeGeneratedSchema().dump(ret))

    @action
    def decode(self, image_file: str) -> dict:
        """
        Decode a QR code from an image file.

        :param image_file: Path of the image file.
        :return: .. schema:: qrcode.QrcodeDecodedSchema
        """
        image_file = os.path.abspath(os.path.expanduser(image_file))
        with open(image_file, 'rb') as f:
            img = Image.open(f)
            results = pyzbar.decode(img)
            return dict(
                QrcodeDecodedSchema().dump(
                    {
                        'results': results,
                        'image_file': image_file,
                    }
                )
            )

    @action
    def start_scanning(
        self,
        camera_plugin: Optional[str] = None,
        duration: Optional[float] = None,
        n_codes: Optional[int] = None,
    ) -> Optional[List[dict]]:
        """
        Decode QR-codes and bar codes using a camera.

        :param camera_plugin: Camera plugin (overrides default ``camera_plugin``).
        :param duration: How long the capturing phase should run (default:
            until ``stop_scanning`` or app termination).
        :param n_codes: Stop after decoding this number of codes (default: None).
        :return: .. schema:: qrcode.QrcodeDecodedResultSchema(many=True)
        """
        assert not self._capturing.is_set(), 'A capturing process is already running'

        camera = self._get_camera(camera_plugin)
        codes = []
        last_results = {}
        last_results_timeout = 5.0
        last_results_time = 0
        self._capturing.set()

        try:
            with camera.open(
                stream=True,
                frames_dir=None,
            ) as session:
                camera.start_camera(session)
                start_time = time.time()

                while (
                    self._capturing.is_set()
                    and (not duration or time.time() < start_time + duration)
                    and (not n_codes or len(codes) < n_codes)
                ):
                    img = camera.capture_frame(session)
                    results = pyzbar.decode(img)

                    if results:
                        results = [
                            result
                            for result in QrcodeDecodedResultSchema().dump(
                                results, many=True
                            )
                            if result['data'] not in last_results
                            or time.time() >= last_results_time + last_results_timeout
                        ]

                    if results:
                        codes.extend(results)
                        get_bus().post(QrcodeScannedEvent(results=results))
                        last_results = {result['data']: result for result in results}
                        last_results_time = time.time()
        finally:
            self._capturing.clear()

        return codes

    @action
    def stop_scanning(self):
        self._capturing.clear()


# vim:sw=4:ts=4:et:



from datetime import datetime
from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, Optional


class DeviceMode(Enum):
    """
    GPS device mode.
    """

    NO_FIX = 1  # No fix
    TWO_D = 2  # 2D fix
    THREE_D = 3  # 3D fix (including altitude)


@dataclass
class GpsDevice:
    """
    Models the GPS device.
    """

    path: str
    activated: Optional[datetime] = None
    native: bool = False
    baudrate: Optional[int] = None
    parity: str = 'N'
    stopbits: Optional[int] = None
    cycle: Optional[float] = None
    driver: Optional[str] = None
    subtype: Optional[str] = None
    mode: Optional[DeviceMode] = None


@dataclass
class GpsStatus:
    """
    Models the status of the GPS service.
    """

    latitude: Optional[float] = None
    longitude: Optional[float] = None
    altitude: Optional[float] = None
    speed: Optional[float] = None
    satellites_used: int = 0
    devices: Dict[str, GpsDevice] = field(default_factory=dict)
    timestamp: Optional[datetime] = None



import os
from typing import Optional, Union, Callable

from platypush.context import get_bus
from platypush.message.event.ngrok import (
    NgrokProcessStartedEvent,
    NgrokTunnelStartedEvent,
    NgrokTunnelStoppedEvent,
    NgrokProcessStoppedEvent,
)
from platypush.plugins import Plugin, action
from platypush.schemas.ngrok import NgrokTunnelSchema


class NgrokPlugin(Plugin):
    """
    Plugin to dynamically create and manage network tunnels using `ngrok <https://ngrok.com/>`_.
    """

    def __init__(
        self,
        auth_token: Optional[str] = None,
        ngrok_bin: Optional[str] = None,
        region: Optional[str] = None,
        **kwargs,
    ):
        """
        :param auth_token: Specify the ``ngrok`` auth token, enabling authenticated features (e.g. more concurrent
            tunnels, custom subdomains, etc.).
        :param ngrok_bin: By default ``pyngrok`` manages its own version of the ``ngrok`` binary, but you can specify
            this option if you want to use a different binary installed on the system.
        :param region: ISO code of the region/country that should host the ``ngrok`` tunnel (default: ``us``).
        """
        from pyngrok import conf, ngrok

        super().__init__(**kwargs)

        conf.get_default().log_event_callback = self._get_event_callback()
        self._active_tunnels_by_url = {}

        if auth_token:
            ngrok.set_auth_token(auth_token)
        if ngrok_bin:
            conf.get_default().ngrok_path = os.path.expanduser(ngrok_bin)
        if region:
            conf.get_default().region = region

    @property
    def _active_tunnels_by_name(self) -> dict:
        return {
            tunnel['name']: tunnel for tunnel in self._active_tunnels_by_url.values()
        }

    def _get_event_callback(self) -> Callable:
        from pyngrok.process import NgrokLog

        def callback(log: NgrokLog):
            if log.msg == 'client session established':
                get_bus().post(NgrokProcessStartedEvent())
            elif log.msg == 'started tunnel':
                tunnel = {
                    'name': log.name,
                    'url': log.url,
                    'protocol': log.url.split(':')[0],
                }

                self._active_tunnels_by_url[tunnel['url']] = tunnel
                get_bus().post(NgrokTunnelStartedEvent(**tunnel))
            elif (
                log.msg == 'end'
                and int(getattr(log, 'status', 0)) == 204
                and getattr(log, 'pg', '').startswith('/api/tunnels')
            ):
                tunnel = log.pg.split('/')[-1]
                tunnel = self._active_tunnels_by_name.pop(
                    tunnel, self._active_tunnels_by_url.pop(tunnel, None)
                )
                if tunnel:
                    get_bus().post(NgrokTunnelStoppedEvent(**tunnel))
            elif log.msg == 'received stop request':
                get_bus().post(NgrokProcessStoppedEvent())

        return callback

    @action
    def create_tunnel(
        self,
        resource: Union[int, str] = 80,
        protocol: str = 'tcp',
        name: Optional[str] = None,
        auth: Optional[str] = None,
        **kwargs,
    ) -> dict:
        """
        Create an ``ngrok`` tunnel to the specified localhost port/protocol.

        :param resource: This can be any of the following:

            - A TCP or UDP port exposed on localhost.
            - A local network address (or ``address:port``) to expose.
            - The absolute path (starting with ``file://``) to a local folder - in such case, the specified directory
              will be served over HTTP through an ``ngrok`` endpoint (see https://ngrok.com/docs#http-file-urls).

            Default: localhost port 80.

        :param protocol: Network protocol (default: ``tcp``).
        :param name: Optional tunnel name.
        :param auth: HTTP basic authentication credentials associated with the tunnel, in the format of
            ``username:password``.
        :param kwargs: Extra arguments supported by the ``ngrok`` tunnel, such as ``hostname``, ``subdomain`` or
            ``remote_addr`` - see the `ngrok documentation <https://ngrok.com/docs#tunnel-definitions>`_ for a full
            list.
        :return: .. schema:: ngrok.NgrokTunnelSchema
        """
        from pyngrok import ngrok

        if isinstance(resource, str) and resource.startswith('file://'):
            protocol = None

        tunnel = ngrok.connect(resource, proto=protocol, name=name, auth=auth, **kwargs)
        return NgrokTunnelSchema().dump(tunnel)

    @action
    def close_tunnel(self, tunnel: str):
        """
        Close an ``ngrok`` tunnel.

        :param tunnel: Name or public URL of the tunnel to be closed.
        """
        from pyngrok import ngrok

        if tunnel in self._active_tunnels_by_name:
            tunnel = self._active_tunnels_by_name[tunnel]['url']

        assert (
            tunnel in self._active_tunnels_by_url
        ), f'No such tunnel URL or name: {tunnel}'
        ngrok.disconnect(tunnel)

    @action
    def get_tunnels(self):
        """
        Get the list of active ``ngrok`` tunnels.

        :return: .. schema:: ngrok.NgrokTunnelSchema(many=True)
        """
        from pyngrok import ngrok

        tunnels = ngrok.get_tunnels()
        return NgrokTunnelSchema().dump(tunnels, many=True)

    @action
    def kill_process(self):
        """
        The first created tunnel instance also starts the ``ngrok`` process.
        The process will stay alive until the Python interpreter is stopped or this action is invoked.
        """
        from pyngrok import ngrok

        proc = ngrok.get_ngrok_process()
        assert proc and proc.proc, 'The ngrok process is not running'
        proc.proc.kill()
        get_bus().post(NgrokProcessStoppedEvent())


# vim:sw=4:ts=4:et:



from dataclasses import dataclass
from typing import Optional

from .._transport import TransportEncryption


@dataclass
class ServerConfig:
    """
    Configuration for a mail server.
    """

    server: str
    port: int
    encryption: TransportEncryption
    timeout: float
    keyfile: Optional[str] = None
    certfile: Optional[str] = None
    domain: Optional[str] = None


# vim:sw=4:ts=4:et:



from typing import Union

import aioxmpp

from platypush.message.event.xmpp import (
    XmppPresenceChangedEvent,
    XmppRoomUserAvailableEvent,
    XmppRoomUserUnavailableEvent,
    XmppUserAvailableEvent,
    XmppUserUnavailableEvent,
)

from .._types import XmppPresence
from ._base import XmppBaseHandler


# pylint: disable=too-many-ancestors
class XmppPresenceHandler(XmppBaseHandler):
    """
    Handler for XMPP presence events.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.presence_client: aioxmpp.PresenceClient = self._client.summon(
            aioxmpp.PresenceClient
        )
        self.presence_client.on_changed.connect(self._on_presence_changed)  # type: ignore
        self.presence_client.on_available.connect(self._on_presence_available)  # type: ignore
        self.presence_client.on_unavailable.connect(self._on_presence_unavailable)  # type: ignore

        self.presence_server: aioxmpp.PresenceServer = self._client.summon(
            aioxmpp.PresenceServer
        )

    def _is_room_event(self, jid: aioxmpp.JID) -> bool:
        jid_str = str(jid.replace(resource=None))  # type: ignore
        return (
            self._state.rooms.get(jid_str) is not None
            or jid_str in self._state.pending_rooms
        )

    def _on_presence_changed(
        self,
        user_id: Union[str, aioxmpp.JID],
        presence: aioxmpp.stanza.Presence,
        **_,
    ):
        if isinstance(user_id, aioxmpp.JID) and self._is_room_event(user_id):
            return  # Rooms will have their own presence changed events

        self._post_user_event(
            XmppPresenceChangedEvent,
            user_id=user_id,
            status=aioxmpp.PresenceShow(presence.show).value
            or XmppPresence.AVAILABLE.value,
        )

    def _on_presence_available(self, user_id: aioxmpp.JID, *_, **__):
        jid = str(user_id)
        if self._is_room_event(user_id):
            self._post_user_event(
                XmppRoomUserAvailableEvent,
                user_id=jid,
                room_id=str(user_id.replace(resource=None)),
                is_self=jid == str(self._jid),
            )
        elif jid in self._state.users:
            self._state.users.add(jid)
            self._post_user_event(XmppUserAvailableEvent, user_id=user_id)

    def _on_presence_unavailable(self, user_id: aioxmpp.JID, *_, **__):
        jid = str(user_id)
        if self._is_room_event(user_id):
            self._post_user_event(
                XmppRoomUserUnavailableEvent,
                user_id=jid,
                room_id=str(user_id.replace(resource=None)),
                is_self=jid == str(self._jid),
            )
        elif jid in self._state.users:
            self._post_user_event(XmppUserUnavailableEvent, user_id=user_id)

    def set_presence(self, presence: XmppPresence):
        available = presence.value != XmppPresence.OFFLINE.value
        presence_show = aioxmpp.PresenceShow(
            None
            if presence.value
            in {XmppPresence.AVAILABLE.value, XmppPresence.OFFLINE.value}
            else presence.value
        )

        self.presence_server.set_presence(
            aioxmpp.PresenceState(
                available=available,
                show=presence_show,
            )
        )



from marshmallow import pre_load

from .._base import SystemBaseSchema


class CpuInfoBaseSchema(SystemBaseSchema):
    """
    Base schema for CPU info.
    """

    @pre_load
    def pre_load(self, data: dict, **_) -> dict:
        data = super().pre_load(data)
        if data.get('hz_advertised'):
            data['frequency_advertised'] = data.pop('hz_advertised')[0]
        if data.get('hz_actual'):
            data['frequency_actual'] = data.pop('hz_actual')[0]

        for key, value in data.items():
            if key.endswith("_cache_size") and isinstance(value, str):
                tokens = value.split(" ")
                unit = None
                if len(tokens) > 1:
                    unit = tokens[1]

                value = int(tokens[0])
                if unit == "KiB":
                    value *= 1024
                elif unit == "MiB":
                    value *= 1024 * 1024
                elif unit == "GiB":
                    value *= 1024 * 1024 * 1024

                data[key] = value

        return data


class CpuTimesBaseSchema(SystemBaseSchema):
    """
    Base schema for CPU times.
    """

    @pre_load
    def pre_load(self, data, **_) -> dict:
        """
        Convert the underlying object to dict and normalize all the percentage
        values from [0, 100] to [0, 1].
        """
        data = super().pre_load(data)
        return {
            key: value / 100.0
            for key, value in (
                data if isinstance(data, dict) else data._asdict()
            ).items()
        }



from typing import Iterable, Mapping, Union

Numeric = Union[float, int]
SensorDataType = Union[Numeric, Mapping[str, Numeric], Iterable[Numeric]]
"""
Numeric sensor data published by integrations can be either of:

    - ``int``/``float``
    - Mapping of ``str -> int/float``
    - List of ``int``/``float``
"""



from marshmallow import EXCLUDE, fields
from marshmallow.schema import Schema


class JoystickStateSchema(Schema):
    """
    Joystick state schema.
    """

    left_joystick_y = fields.Float(
        metadata={
            'description': 'Left joystick Y axis value',
            'example': 0.5,
        }
    )

    left_joystick_x = fields.Float(
        metadata={
            'description': 'Left joystick X axis value',
            'example': 0.5,
        }
    )

    right_joystick_y = fields.Float(
        metadata={
            'description': 'Right joystick Y axis value',
            'example': 0.5,
        }
    )

    right_joystick_x = fields.Float(
        metadata={
            'description': 'Right joystick X axis value',
            'example': 0.5,
        }
    )

    left_trigger = fields.Float(
        metadata={
            'description': 'Left trigger value',
            'example': 0.5,
        }
    )

    right_trigger = fields.Float(
        metadata={
            'description': 'Right trigger value',
            'example': 0.5,
        }
    )

    left_bumper = fields.Integer(
        metadata={
            'description': 'Left bumper state',
            'example': 1,
        }
    )

    right_bumper = fields.Integer(
        metadata={
            'description': 'Right bumper state',
            'example': 1,
        }
    )

    a = fields.Integer(
        metadata={
            'description': 'A button state',
            'example': 1,
        }
    )

    x = fields.Integer(
        metadata={
            'description': 'X button state',
            'example': 1,
        }
    )

    y = fields.Integer(
        metadata={
            'description': 'Y button state',
            'example': 1,
        }
    )

    b = fields.Integer(
        metadata={
            'description': 'B button state',
            'example': 1,
        }
    )

    left_thumb = fields.Integer(
        metadata={
            'description': 'Left thumb button state',
            'example': 1,
        }
    )

    right_thumb = fields.Integer(
        metadata={
            'description': 'Right thumb button state',
            'example': 1,
        }
    )

    back = fields.Integer(
        metadata={
            'description': 'Back button state',
            'example': 1,
        }
    )

    start = fields.Integer(
        metadata={
            'description': 'Start button state',
            'example': 1,
        }
    )

    left_dir_pad = fields.Integer(
        metadata={
            'description': 'Left direction pad button state',
            'example': 1,
        }
    )

    right_dir_pad = fields.Integer(
        metadata={
            'description': 'Right direction pad button state',
            'example': 1,
        }
    )

    up_dir_pad = fields.Integer(
        metadata={
            'description': 'Up direction pad button state',
            'example': 1,
        }
    )

    down_dir_pad = fields.Integer(
        metadata={
            'description': 'Down direction pad button state',
            'example': 1,
        }
    )


class JoystickDeviceSchema(Schema):
    """
    Joystick device schema.
    """

    class Meta:
        """
        Meta class.
        """

        unknown = EXCLUDE

    name = fields.String(
        metadata={
            'description': 'Joystick name',
            'example': 'Xbox 360 Controller',
        }
    )

    path = fields.Function(
        lambda obj: obj.get_char_device_path(),
        metadata={
            'description': 'Joystick character device path',
            'example': '/dev/input/event0',
        },
    )

    number = fields.Integer(
        metadata={
            'description': 'Joystick number',
            'example': 0,
        }
    )

    protocol = fields.String(
        metadata={
            'description': 'Joystick protocol',
            'example': 'usb',
        }
    )


class JoystickStatusSchema(Schema):
    """
    Joystick status schema.
    """

    device = fields.Nested(JoystickDeviceSchema, required=True)
    state = fields.Nested(JoystickStateSchema, required=True)



import os
import pytest
import tempfile
import time

from platypush.message.event.custom import CustomEvent

from .utils import register_user, send_request

tmp_files = []


@pytest.fixture(scope='module', autouse=True)
def user(*_):
    register_user()


@pytest.fixture(scope='module', autouse=True)
def tmp_file(*_):
    n_files = 2
    for _ in range(n_files):
        tmp_file = tempfile.NamedTemporaryFile(prefix='platypush-test-procedure-', suffix='.txt', delete=False)
        tmp_files.append(tmp_file.name)

    yield

    for f in tmp_files:
        if os.path.isfile(f):
            os.unlink(f)


def check_file_content(expected_content: str, tmp_file: str, timeout: int = 10):
    error = None
    start_time = time.time()

    while time.time() - start_time < timeout:
        try:
            assert os.path.isfile(tmp_file), 'The expected output file was not created'
            with open(tmp_file, 'r') as f:
                content = f.read()

            assert content == expected_content, 'The output file did not contain the expected text'
            error = None
            break
        except Exception as e:
            error = e

    if error:
        raise error


def test_procedure_call(base_url):
    """
    Test the result of a procedure invoked directly over HTTP.
    """
    output_text = 'Procedure test'
    send_request(
        action='procedure.write_file',
        args={
            'file': tmp_files[0],
            'content': output_text,
        })

    check_file_content(expected_content=output_text, tmp_file=tmp_files[0])


def test_procedure_from_event(app, base_url):
    """
    Test the result of a procedure triggered by an event.
    """
    output_text = 'Procedure from event test'
    event_type = 'test_procedure'
    # noinspection PyUnresolvedReferences
    app.bus.post(CustomEvent(subtype=event_type, file=tmp_files[1], content=output_text))
    check_file_content(expected_content=output_text, tmp_file=tmp_files[1])


if __name__ == '__main__':
    pytest.main()


# vim:sw=4:ts=4:et:



import importlib
import inspect
import logging
import os
import pkgutil
from typing import List

from platypush.plugins.bluetooth._plugins import BaseBluetoothPlugin

logger = logging.getLogger(__name__)


def scan_plugins(manager) -> List[BaseBluetoothPlugin]:
    """
    Initializes all the plugins associated to the given BluetoothManager by
    scanning all the modules under the ``_plugins`` folder in the manager's
    package.
    """
    plugins = {}
    base_dir = os.path.dirname(inspect.getfile(manager.__class__))
    module = inspect.getmodule(manager.__class__)
    assert module is not None
    package = module.__package__
    assert package is not None

    for _, mod_name, _ in pkgutil.walk_packages([base_dir], prefix=package + '.'):
        try:
            module = importlib.import_module(mod_name)
        except Exception as e:
            logger.warning('Could not import module %s', mod_name)
            logger.exception(e)
            continue

        for _, obj in inspect.getmembers(module):
            if (
                inspect.isclass(obj)
                and not inspect.isabstract(obj)
                and issubclass(obj, BaseBluetoothPlugin)
            ):
                plugins[obj] = obj(manager)

    return list(plugins.values())



import asyncio
from contextlib import asynccontextmanager
import threading
from typing import (
    AsyncGenerator,
    Collection,
    Final,
    List,
    Optional,
    Dict,
    Union,
)

from bleak import BleakClient, BleakScanner
from bleak.backends.device import BLEDevice
from bleak.exc import BleakError

from platypush.context import get_or_create_event_loop
from platypush.entities.bluetooth import BluetoothDevice
from platypush.message.event.bluetooth import (
    BluetoothConnectionFailedEvent,
    BluetoothDeviceDisconnectedEvent,
    BluetoothDeviceLostEvent,
)

from .._manager import BaseBluetoothManager
from .._types import RawServiceClass
from ._cache import DeviceCache
from ._connection import BluetoothConnection
from ._event_handler import EventHandler


class BLEManager(BaseBluetoothManager):
    """
    Integration for Bluetooth Low Energy (BLE) devices.
    """

    _rssi_update_interval: Final[int] = 30
    """
    How long we should wait before triggering an update event upon a new
    RSSI update, in seconds.
    """

    _min_get_device_scan_duration: Final[int] = 10
    """
    Wait at least this many seconds for a device to be discovered when calling
    get_device().
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._connections: Dict[str, BluetoothConnection] = {}
        """
        ``address -> BluetoothConnection`` mapping containing the active
        connections
        """
        self._connection_locks: Dict[str, asyncio.Lock] = {}
        """
        ``address -> Lock`` locks used to synchronize concurrent access to
        the devices
        """
        self._device_cache = DeviceCache()
        """ Cache of discovered ``BLEDevice`` objects. """
        self._event_handler = EventHandler(
            device_queue=self._device_queue,
            device_cache=self._device_cache,
            entity_cache=self._cache,
            plugins=self._plugins,
            blacklist=self._blacklist,
            exclude_known_noisy_beacons=self._exclude_known_noisy_beacons,
        )
        """ Bluetooth device event handler """
        self._main_loop: Optional[asyncio.AbstractEventLoop] = None
        """ Main event loop """

    async def _get_device(self, device: str) -> BLEDevice:
        """
        Utility method to get a device by name or address.
        """
        dev = self._device_cache.get(device)
        if not dev:
            self.logger.info('Scanning for unknown device "%s"', device)
            await self._scan(
                duration=max(self._min_get_device_scan_duration, self.poll_interval)
            )

            dev = self._device_cache.get(device)

        assert dev, f'Unknown device: "{device}"'
        return dev

    def _disconnected_callback(self, client: BleakClient):
        self._connections.pop(client.address, None)
        dev = self._cache.get(client.address)
        if not dev:
            return  # Unknown device

        dev.connected = False
        self.notify(BluetoothDeviceDisconnectedEvent, dev)

    @asynccontextmanager
    async def _connect(
        self,
        device: str,
        interface: Optional[str] = None,
        timeout: Optional[float] = None,
        close_event: Optional[asyncio.Event] = None,
    ) -> AsyncGenerator[BluetoothConnection, None]:
        """
        Asynchronous context manager that wraps a BLE device connection.
        """
        dev = await self._get_device(device)

        async with self._connection_locks.get(dev.address, asyncio.Lock()) as lock:
            self._connection_locks[dev.address] = lock or asyncio.Lock()

            try:
                async with BleakClient(
                    dev.address,
                    adapter=interface or self._interface,
                    timeout=timeout or self._connect_timeout,
                    disconnected_callback=self._disconnected_callback,
                ) as client:
                    entity = self._cache.get(client.address)
                    if not client:
                        if entity:
                            entity.connected = False
                            self.notify(BluetoothConnectionFailedEvent, entity)

                        raise AssertionError(
                            f'Could not connect to the device {device}'
                        )

                    # Yield the BluetoothConnection object
                    self._connections[dev.address] = BluetoothConnection(
                        client=client,
                        device=dev,
                        loop=asyncio.get_event_loop(),
                        thread=threading.current_thread(),
                        close_event=close_event,
                    )
                    yield self._connections[dev.address]
            except BleakError as e:
                raise AssertionError(f'Could not connect to the device {device}') from e

    async def _read(
        self,
        device: str,
        service_uuid: RawServiceClass,
        interface: Optional[str] = None,
        connect_timeout: Optional[float] = None,
    ) -> bytearray:
        """
        Asynchronously read the next chunk of raw bytes from a BLE device given
        a service UUID.
        """
        async with self._connect(device, interface, connect_timeout) as conn:
            try:
                data = await conn.client.read_gatt_char(service_uuid)
            except BleakError as e:
                raise AssertionError(
                    f'Could not read from the device {device}: {e}'
                ) from e

        return data

    async def _write(
        self,
        device: str,
        data: bytes,
        service_uuid: RawServiceClass,
        interface: Optional[str] = None,
        connect_timeout: Optional[float] = None,
    ):
        """
        Asynchronously write a chunk of raw bytes to a BLE device given a
        service UUID.
        """
        async with self._connect(device, interface, connect_timeout) as conn:
            try:
                await conn.client.write_gatt_char(service_uuid, data)
            except BleakError as e:
                raise AssertionError(
                    f'Could not write to the device {device}: {e}'
                ) from e

    async def _scan(
        self,
        duration: Optional[float] = None,
        service_uuids: Optional[Collection[RawServiceClass]] = None,
    ) -> List[BluetoothDevice]:
        """
        Asynchronously scan for BLE devices and return the discovered devices
        as a list of :class:`platypush.entities.bluetooth.BluetoothDevice`
        entities.
        """
        with self._scan_lock:
            timeout = duration or self.poll_interval
            devices = await BleakScanner.discover(
                adapter=self._interface,
                timeout=timeout,
                service_uuids=list(
                    map(str, service_uuids or self._service_uuids or [])
                ),
                detection_callback=self._event_handler,
            )

        addresses = {dev.address.lower() for dev in devices}
        return [
            dev
            for addr, dev in self._cache.items()
            if addr.lower() in addresses and dev.reachable
        ]

    def _close_active_connections(self):
        """
        Terminates all active connections.
        """
        connections = list(self._connections.values())
        for conn in connections:
            try:
                self.disconnect(conn.device.address)
            except Exception as e:
                self.logger.warning(
                    'Error while disconnecting from %s: %s', conn.device.address, e
                )

    def _dbus_disconnect(self, device: BLEDevice):
        """
        Disconnect from a device using the DBus API (only available on Linux).
        This may be required if there is an active connection to the device that is not owned by the
        """
        entity = self._cache.get(device.address)
        assert entity, f'Unknown device: "{device.address}"'
        path = device.details.get('path')
        assert path, f'The device "{device.address}" has no reported system path'
        assert path.startswith('/org/bluez/'), (
            f'The device "{device.address}" is not a BlueZ device. Programmatic '
            'system disconnection is only supported on Linux'
        )

        try:
            import pydbus

            bus = pydbus.SystemBus()
            dbus_device = bus.get('org.bluez', path)
            dbus_device.Disconnect()
        except Exception as e:
            raise AssertionError(
                f'Could not disconnect from {device.address}: {e}'
            ) from e

        if entity.connected:
            entity.connected = False
            self.notify(BluetoothDeviceDisconnectedEvent, entity)

    def connect(
        self,
        device: str,
        port: Optional[int] = None,
        service_uuid: Optional[RawServiceClass] = None,
        interface: Optional[str] = None,
        timeout: Optional[float] = None,
    ):
        timeout = timeout or self._connect_timeout
        connected_event = threading.Event()
        close_event = asyncio.Event()
        loop = asyncio.new_event_loop()

        def connect_thread():
            """
            The connection thread. It wraps an asyncio loop with a connect
            context manager.
            """

            async def connect_wrapper():
                """
                The asyncio connect wrapper.
                """
                async with self._connect(device, interface, timeout, close_event):
                    connected_event.set()
                    await close_event.wait()

            asyncio.set_event_loop(loop)
            loop.run_until_complete(connect_wrapper())

        # Initialize the loop and start the connection thread
        loop = get_or_create_event_loop()
        connector = threading.Thread(
            target=connect_thread,
            name=f'Bluetooth:connect@{device}',
        )
        connector.start()

        # Wait for the connection to succeed
        success = connected_event.wait(timeout=timeout)
        assert success, f'Connection to {device} timed out'

    def disconnect(self, device: str, *_, **__):
        # Get the device
        loop = get_or_create_event_loop()
        dev = loop.run_until_complete(self._get_device(device))
        assert dev, f'Device {device} not found'

        # Check if there are any active connections
        connection = self._connections.get(dev.address, None)
        if not connection:
            # If there are no active connections in this process, try to
            # disconnect through the DBus API
            self._dbus_disconnect(dev)
            return

        # Set the close event and wait for any connection thread to terminate
        if connection.close_event:
            connection.close_event.set()
        if connection.thread and connection.thread.is_alive():
            connection.thread.join(timeout=5)
        assert not (
            connection.thread and connection.thread.is_alive()
        ), f'Disconnection from {device} timed out'

    def scan(
        self,
        duration: Optional[float] = None,
        service_uuids: Optional[Collection[RawServiceClass]] = None,
    ) -> List[BluetoothDevice]:
        """
        Scan for Bluetooth devices nearby and return the results as a list of
        entities.

        :param duration: Scan duration in seconds (default: same as the plugin's
            `poll_interval` configuration parameter)
        :param service_uuids: List of service UUIDs to discover. Default: any.
        """
        loop = get_or_create_event_loop()
        return loop.run_until_complete(self._scan(duration, service_uuids))

    def read(
        self,
        device: str,
        service_uuid: RawServiceClass,
        interface: Optional[str] = None,
        connect_timeout: Optional[float] = None,
    ) -> bytearray:
        """
        :param device: Name or address of the device to read from.
        :param service_uuid: Service UUID.
        :param interface: Bluetooth adapter name to use (default configured if None).
        :param connect_timeout: Connection timeout in seconds (default: same as the
            configured `connect_timeout`).
        """
        loop = get_or_create_event_loop()
        return loop.run_until_complete(
            self._read(device, service_uuid, interface, connect_timeout)
        )

    def write(
        self,
        device: str,
        data: Union[bytes, bytearray],
        service_uuid: RawServiceClass,
        interface: Optional[str] = None,
        connect_timeout: Optional[float] = None,
    ):
        get_or_create_event_loop().run_until_complete(
            self._write(device, data, service_uuid, interface, connect_timeout)
        )

    async def listen(self):
        """
        Main loop listener.
        """
        self.logger.info('Starting BLE scanner')
        device_addresses = set()

        while not self.should_stop():
            self._scan_enabled.wait()
            if self.should_stop():
                break

            entities = await self._scan(service_uuids=self._service_uuids)
            new_device_addresses = {e.external_id for e in entities}
            missing_device_addresses = device_addresses - new_device_addresses
            missing_devices = [
                dev
                for addr, dev in self._cache.items()
                if addr in missing_device_addresses
            ]

            for dev in missing_devices:
                dev.reachable = False
                dev.connected = False
                self.notify(BluetoothDeviceLostEvent, dev)

            device_addresses = new_device_addresses

    def run(self):
        super().run()

        self._main_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self._main_loop)
        try:
            self._main_loop.run_until_complete(self.listen())
        except Exception as e:
            if not self.should_stop():
                self.logger.warning('The main loop failed unexpectedly: %s', e)
                self.logger.exception(e)
        finally:
            try:
                # Try and release the scan lock if acquired
                self._scan_lock.release()
            except Exception:
                pass

    def stop(self):
        """
        Upon stop request, it stops any pending scans and closes all active
        connections.
        """
        super().stop()
        self._close_active_connections()
        if self._main_loop and self._main_loop.is_running():
            self._main_loop.stop()

        self.logger.info('Stopped the BLE scanner')


# vim:sw=4:ts=4:et:



import logging

from telegram import Message as TelegramMessage, User as TelegramUser

from platypush.schemas.telegram import (
    TelegramMessageSchema,
    TelegramUserSchema,
)

log = logging.getLogger(__name__)


def dump_msg(msg: TelegramMessage) -> dict:
    return dict(TelegramMessageSchema().dump(msg))


def dump_user(user: TelegramUser) -> dict:
    return dict(TelegramUserSchema().dump(user))


# vim:sw=4:ts=4:et:



import json

from marshmallow import EXCLUDE, fields, pre_dump
from marshmallow.schema import Schema

from platypush.schemas import DateTime


class PushbulletActionSchema(Schema):
    """
    Schema for Pushbullet notification actions.
    """

    label = fields.String(
        required=True,
        metadata={
            'description': 'Label of the action',
            'example': 'Example action',
        },
    )

    trigger_key = fields.String(
        required=True,
        metadata={
            'description': 'Key of the action',
            'example': 'example_action',
        },
    )


class PushbulletSchema(Schema):
    """
    Schema for Pushbullet API messages.
    """

    # pylint: disable=too-few-public-methods
    class Meta:  # type: ignore
        """
        Exclude unknown fields.
        """

        unknown = EXCLUDE

    notification_id = fields.String(
        required=True,
        metadata={
            'description': 'Unique identifier for the notification/message',
            'example': '12345',
        },
    )

    title = fields.String(
        metadata={
            'description': 'Title of the notification/message',
            'example': 'Hello world',
        },
    )

    body = fields.Raw(
        metadata={
            'description': 'Body of the notification/message',
            'example': 'Example body',
        },
    )

    url = fields.Url(
        metadata={
            'description': 'URL attached to the notification/message',
            'example': 'https://example.com',
        },
    )

    source_user = fields.String(
        attribute='source_user_iden',
        metadata={
            'description': 'Source user of the notification/message',
            'example': 'user123',
        },
    )

    source_device = fields.String(
        attribute='source_device_iden',
        metadata={
            'description': 'Source device of the notification/message',
            'example': 'device123',
        },
    )

    target_device = fields.String(
        attribute='target_device_iden',
        metadata={
            'description': 'Target device of the notification/message',
            'example': 'device456',
        },
    )

    sender_id = fields.String(
        attribute='sender_iden',
        metadata={
            'description': 'Sender ID of the notification/message',
            'example': '12345',
        },
    )

    sender_email = fields.Email(
        attribute='sender_email_normalized',
        metadata={
            'description': 'Sender email of the notification/message',
            'example': 'user1@example.com',
        },
    )

    sender_name = fields.String(
        metadata={
            'description': 'Sender name of the notification/message',
            'example': 'John Doe',
        },
    )

    receiver_id = fields.String(
        attribute='receiver_iden',
        metadata={
            'description': 'Receiver ID of the notification/message',
            'example': '12346',
        },
    )

    receiver_email = fields.Email(
        attribute='receiver_email_normalized',
        metadata={
            'description': 'Receiver email of the notification/message',
            'example': 'user2@example.com',
        },
    )

    dismissible = fields.Boolean(
        metadata={
            'description': 'Whether the notification/message is dismissible',
            'example': True,
        },
    )

    icon = fields.String(
        metadata={
            'description': 'Base64 encoded icon of the notification/message',
            'example': 'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAABpUlEQVQ4T',
        },
    )

    application_name = fields.String(
        metadata={
            'description': 'Name of the application that sent the notification/message',
            'example': 'Example app',
        },
    )

    package_name = fields.String(
        metadata={
            'description': 'Package name of the application that sent the notification/message',
            'example': 'com.example.app',
        },
    )

    file_name = fields.String(
        metadata={
            'description': 'Name of the file attached to the notification/message',
            'example': 'example.txt',
        },
    )

    file_type = fields.String(
        metadata={
            'description': 'Type of the file attached to the notification/message',
            'example': 'text/plain',
        },
    )

    file_url = fields.Url(
        metadata={
            'description': 'URL of the file attached to the notification/message',
            'example': 'https://example.com/example.txt',
        },
    )

    image_width = fields.Integer(
        metadata={
            'description': 'Width of the image attached to the notification/message',
            'example': 100,
        },
    )

    image_height = fields.Integer(
        metadata={
            'description': 'Height of the image attached to the notification/message',
            'example': 100,
        },
    )

    image_url = fields.Url(
        metadata={
            'description': 'URL of the image attached to the notification/message',
            'example': 'https://example.com/example.png',
        },
    )

    actions = fields.Nested(
        PushbulletActionSchema,
        many=True,
        metadata={
            'description': 'Actions of the notification/message',
        },
    )

    created = DateTime(
        metadata={
            'description': 'Creation timestamp of the notification/message',
            'example': '2021-01-01T00:00:00',
        },
    )

    modified = DateTime(
        metadata={
            'description': 'Last modification timestamp of the notification/message',
            'example': '2021-01-01T00:00:00',
        },
    )

    @pre_dump
    def pre_dump(self, data, **_):
        """
        Pre-dump hook.
        """

        data['notification_id'] = str(
            data.pop('iden', data.pop('notification_id', None))
        )

        if data.get('body') is not None:
            try:
                data['body'] = json.loads(data['body'])
            except (TypeError, ValueError):
                pass

        return data


class PushbulletDeviceSchema(Schema):
    """
    Schema for Pushbullet devices.
    """

    active = fields.Boolean(
        metadata={
            'description': 'Whether the device is active',
            'example': True,
        },
    )

    device_id = fields.String(
        required=True,
        attribute='iden',
        metadata={
            'description': 'Unique identifier for the device',
            'example': '12345',
        },
    )

    name = fields.String(
        attribute='nickname',
        metadata={
            'description': 'Name of the device',
            'example': 'Example device',
        },
    )

    kind = fields.String(
        metadata={
            'description': 'Kind of the device',
            'example': 'android',
        },
    )

    manufacturer = fields.String(
        metadata={
            'description': 'Manufacturer of the device',
            'example': 'Example manufacturer',
        },
    )

    model = fields.String(
        metadata={
            'description': 'Model of the device',
            'example': 'Example model',
        },
    )

    icon = fields.String(
        metadata={
            'description': 'Device icon type',
            'example': 'system',
        },
    )

    pushable = fields.Boolean(
        metadata={
            'description': 'Whether it is possible to push notifications and '
            'messages to the device',
            'example': True,
        },
    )

    created = DateTime(
        metadata={
            'description': 'Creation timestamp of the device',
            'example': '2021-01-01T00:00:00',
        },
    )

    modified = DateTime(
        metadata={
            'description': 'Last modification timestamp of the device',
            'example': '2021-01-01T00:00:00',
        },
    )



from abc import ABCMeta
from datetime import datetime
from typing import Union, Optional, Iterable

from dateutil.tz import gettz

from platypush.message.event import Event


class SlackEvent(Event, ABCMeta):
    """
    Base class for Slack events.
    """
    def __init__(self, *args, timestamp: Optional[Union[int, float, datetime]] = None, **kwargs):
        """
        :param timestamp: Event timestamp.
        """
        kwargs['event_time'] = self._convert_timestamp(timestamp)
        super().__init__(*args, **kwargs)

    @staticmethod
    def _convert_timestamp(timestamp: Optional[Union[int, float, datetime]] = None) -> Optional[datetime]:
        if not (isinstance(timestamp, int) or isinstance(timestamp, float)):
            return timestamp

        return datetime.fromtimestamp(timestamp, tz=gettz())   # lgtm [py/call-to-non-callable]


class SlackMessageEvent(SlackEvent, ABCMeta):  # lgtm [py/conflicting-attributes]
    """
    Base class for message-related events.
    """
    def __init__(self, *args, text: str, user: str, channel: Optional[str] = None, team: Optional[str] = None,
                 icons: dict = None, blocks: Iterable[dict] = None, **kwargs):
        """
        :param text: Message text.
        :param user: ID of the sender.
        :param channel: ID of the channel.
        :param team: ID of the team.
        :param icons: Mapping of the icons for this message.
        :param blocks: Extra blocks in the message.
        """
        super().__init__(*args, text=text, user=user, channel=channel, team=team, icons=icons, blocks=blocks, **kwargs)


class SlackMessageReceivedEvent(SlackMessageEvent):
    """
    Event triggered when a message is received on a monitored resource.
    """


class SlackMessageEditedEvent(SlackMessageEvent):
    """
    Event triggered when a message is edited on a monitored resource.
    """


class SlackMessageDeletedEvent(SlackMessageEvent):
    """
    Event triggered when a message is deleted from a monitored resource.
    """


class SlackAppMentionReceivedEvent(SlackMessageEvent):
    """
    Event triggered when a message that mentions the app is received on a monitored resource.
    """


# vim:sw=4:ts=4:et:



import json
import os
import re
import subprocess
import threading
import time
from dataclasses import asdict, dataclass
from multiprocessing import Process, Queue, RLock
from queue import Empty
from typing import Any, Collection, Dict, List, Optional

from platypush.message.response import Response
from platypush.plugins import action
from platypush.plugins.media import PlayerState, MediaPlugin
from platypush.message.event.media import (
    MediaPlayEvent,
    MediaPlayRequestEvent,
    MediaPauseEvent,
    MediaResumeEvent,
    MediaStopEvent,
    NewPlayingMediaEvent,
)
from platypush.utils import find_bins_in_path


@dataclass
class MplayerStatus:
    """
    MPlayer status object
    """

    state: PlayerState = PlayerState.STOP
    filename: Optional[str] = None
    path: Optional[str] = None
    title: Optional[str] = None
    duration: Optional[float] = None
    position: Optional[float] = None
    percent_pos: Optional[float] = None
    fullscreen: Optional[bool] = None
    mute: Optional[bool] = None
    pause: Optional[bool] = None
    volume: Optional[float] = None
    volume_max: Optional[float] = None
    seekable: Optional[bool] = None
    url: Optional[str] = None


class MediaMplayerPlugin(MediaPlugin):
    """
    Plugin to control MPlayer instances.

    Note that some plugin methods are populated dynamically by introspecting the
    mplayer executable. You can verify the supported methods at runtime by
    running the :meth:`.list_actions` action.
    """

    _mplayer_default_communicate_timeout = 0.5

    _mplayer_bin_default_args = [
        '-slave',
        '-quiet',
        '-idle',
        '-input',
        'nodefault-bindings',
        '-noconfig',
        'all',
    ]

    def __init__(
        self,
        fullscreen: bool = False,
        mplayer_bin: Optional[str] = None,
        mplayer_timeout: float = _mplayer_default_communicate_timeout,
        args: Optional[Collection[str]] = None,
        **kwargs,
    ):
        """
        :param fullscreen: If set to True then the player will be started in
            fullscreen mode (default: False)
        :param mplayer_bin: Path to the MPlayer executable (default: search for
            the first occurrence in your system PATH environment variable)
        :param mplayer_timeout: Timeout in seconds to wait for more data
            from MPlayer before considering a response ready (default: 0.5 seconds)
        :param args: Default arguments that will be passed to the MPlayer
            executable
        """

        super().__init__(**kwargs)

        self.args = args or []
        self._init_mplayer_bin(mplayer_bin=mplayer_bin)
        self._fullscreen = fullscreen
        self._build_actions()
        self._player = None
        self._mplayer_timeout = mplayer_timeout
        self._status_lock = threading.Lock()
        self._status = MplayerStatus()
        self._answer_queue = Queue()
        self._proc_monitor: Optional[Process] = None
        self._cmd_lock = RLock()
        self._cleanup_lock = RLock()

    def _init_mplayer_bin(self, mplayer_bin=None):
        if not mplayer_bin:
            bin_name = 'mplayer.exe' if os.name == 'nt' else 'mplayer'
            bins = find_bins_in_path(bin_name)

            if not bins:
                raise RuntimeError(
                    'MPlayer executable not specified and not '
                    + 'found in your PATH. Make sure that mplayer'
                    + 'is either installed or configured'
                )

            self.mplayer_bin = bins[0]
        else:
            mplayer_bin = os.path.expanduser(mplayer_bin)
            if not (
                os.path.isfile(mplayer_bin)
                and (os.name == 'nt' or os.access(mplayer_bin, os.X_OK))
            ):
                raise RuntimeError(
                    f'{mplayer_bin} is does not exist or is not a valid executable file'
                )

            self.mplayer_bin = mplayer_bin

    def _init_mplayer(self, mplayer_args=None):
        if self._player:
            try:
                self._player.terminate()
            except Exception as e:
                self.logger.debug('Failed to quit mplayer before _exec: %s', e)

        m_args = mplayer_args or []
        args = [self.mplayer_bin] + self._mplayer_bin_default_args
        if self._fullscreen and '-fs' not in args:
            args.append('-fs')

        for arg in (*self.args, *m_args):
            if arg not in args:
                args.append(arg)

        popen_args: Dict[str, Any] = {
            'stdin': subprocess.PIPE,
            'stdout': subprocess.PIPE,
        }

        if self._env:
            popen_args['env'] = self._env

        self._player = subprocess.Popen(args, **popen_args)
        self._proc_monitor = Process(target=self._listener, name='mplayer-monitor')
        self._proc_monitor.start()

    def _build_actions(self):
        """Populates the actions list by introspecting the mplayer executable"""

        def args_pprint(txt):
            lc = txt.lower()
            if lc[0] == '[':
                return f'{lc[1:-1]}=None'
            return lc

        self._actions = {}
        with subprocess.Popen(
            [self.mplayer_bin, '-input', 'cmdlist'],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
        ) as mplayer:
            while True:
                if not mplayer.stdout:
                    break

                line = mplayer.stdout.readline()
                if not line:
                    break

                line = line.decode()
                if line[0].isupper():
                    continue

                args = line.split()
                cmd_name = args.pop(0)
                arguments = ', '.join([args_pprint(a) for a in args])
                self._actions[cmd_name] = f'{cmd_name}({arguments})'

    def _exec(
        self, cmd, *args, mplayer_args=None, prefix=None, wait_for_response=False
    ) -> Optional[dict]:
        cmd_name = cmd
        if cmd_name in {'loadfile', 'loadlist'}:
            self._init_mplayer(mplayer_args)
        else:
            if not self._player:
                self.logger.warning('MPlayer is not running')

        cmd = (
            f'{prefix + " " if prefix else ""}'
            + cmd_name
            + (" " if args else "")
            + " ".join(repr(a) for a in args)
            + '\n'
        ).encode()

        if not self._player:
            self.logger.debug('Cannot send command %s: player unavailable', cmd)
            return None

        if not (self._player.stdin and self._player.stdin.writable()):
            self.logger.warning(
                'Could not communicate with the mplayer process: the stdin is closed'
            )
            return None

        # Make sure that the response queue is empty before waiting for a new response
        while not self._answer_queue.empty():
            self._answer_queue.get()

        self.logger.debug('mplayer interface:: Sending command: %s', cmd)

        with self._cmd_lock:
            try:
                self._player.stdin.write(cmd)
                self._player.stdin.flush()
            except BrokenPipeError:
                self.logger.info('The MPlayer process has terminated')
                self._cleanup()
                return None
            except Exception as e:
                self.logger.warning(
                    'Failed to send command %s: %s: %s', cmd, type(e).__name__, e
                )
                return None

        if cmd_name in {'loadfile', 'loadlist'}:
            self._post_event(NewPlayingMediaEvent, resource=args[0])

        if not wait_for_response:
            return None

        # Get the response from the queue
        try:
            ret, status = self._answer_queue.get(
                block=True, timeout=self._mplayer_timeout
            )
            self._status = status
        except Empty:
            self.logger.warning('No response from mplayer for command %s', cmd)
            return None

        return ret

    def _process_answer(self, answer: dict):
        for k, v in answer.items():
            if k == 'pause':
                if v and self._status.state == PlayerState.PLAY:
                    self._status.state = PlayerState.PAUSE
                    self._post_event(MediaPauseEvent)
                elif not v:
                    if self._status.state == PlayerState.PAUSE:
                        self._post_event(MediaResumeEvent)
                    elif self._status.state == PlayerState.STOP:
                        self._post_event(MediaPlayEvent)

                    self._status.state = PlayerState.PLAY
            elif k == 'filename':
                self._status.filename = v
            elif k == 'path':
                self._status.path = v
                self._status.url = (
                    'file://' if os.path.isfile(v) else ''
                ) + self._status.path
            elif k == 'fullscreen':
                self._status.fullscreen = v
            elif k == 'mute':
                self._status.mute = v
            elif k == 'percent_pos':
                self._status.percent_pos = v
            elif k == 'time_pos':
                self._status.position = v
            elif k == 'volume':
                self._status.volume = v
            elif k == 'length':
                self._status.duration = v

        self._answer_queue.put((answer, self._status))

    def _status_checker(self):
        try:
            while self._player and self._player.stdin and self._player.stdin.writable():
                try:
                    self._get_property('filename')
                except (IOError, ValueError, KeyboardInterrupt):
                    break
                finally:
                    time.sleep(1)
        except Exception as e:
            self.logger.warning('mplayer status checker process failed: %s', e)

    def _listener(self):
        status_checker = Process(
            target=self._status_checker, name='mplayer-status-checker'
        )

        try:
            status_checker.start()

            while (
                self._player and self._player.stdout and self._player.stdout.readable()
            ):
                try:
                    buf = self._player.stdout.readline()
                except (IOError, ValueError, KeyboardInterrupt):
                    break

                line = buf.decode() if isinstance(buf, bytes) else buf
                self.logger.debug('mplayer interface:: Received line: %s', buf)

                if not line:
                    break

                if line.startswith('ANS_'):
                    m = re.match(r'^([^=]+)=(.*)\s*$', line[4:])
                    if not m:
                        self.logger.warning('Unexpected response: %s', line)
                        break

                    k, v = m.group(1), m.group(2)
                    if v == 'yes':
                        v = True
                    elif v == 'no':
                        v = False

                    try:
                        if isinstance(v, str):
                            v = json.loads(v)
                    except (TypeError, ValueError):
                        pass

                    self._process_answer({k: v})
                elif line.startswith('Starting playback'):
                    self._status.state = PlayerState.PLAY
                    self._post_event(MediaPlayEvent)
        finally:
            if status_checker and status_checker.is_alive():
                status_checker.terminate()
                status_checker.join(timeout=5)
                try:
                    status_checker.kill()
                except Exception:
                    pass

            if self._player:
                self._player.wait()
                try:
                    self.quit()
                except Exception:
                    pass

            self._player = None

    @action
    def execute(self, cmd, args=None):
        """
        Execute a raw MPlayer command. See
        https://www.mplayerhq.hu/DOCS/tech/slave.txt for a reference or call
        :meth:`.list_actions` to get a list
        """

        args = args or []
        return self._exec(cmd, *args)

    @action
    def list_actions(self):
        return [
            {'action': a, 'args': self._actions[a]}
            for a in sorted(self._actions.keys())
        ]

    def _post_event(self, evt_type, **evt):
        self._bus.post(
            evt_type(
                player='local',
                plugin='media.mplayer',
                resource=evt.pop('resource', self._status.url),
                title=self._status.title or self._status.filename,
                **evt,
            )
        )

    @action
    def play(
        self,
        resource: str,
        subtitles: Optional[str] = None,
        mplayer_args: Optional[List[str]] = None,
        **_,
    ):
        """
        Play a resource.

        :param resource: Resource to play - can be a local file or a remote URL
        :param subtitles: Path to optional subtitle file
        :param mplayer_args: Extra runtime arguments that will be passed to the
            MPlayer executable
        """

        self._post_event(MediaPlayRequestEvent, resource=resource)
        if subtitles:
            subs = self.get_subtitles_file(subtitles)
            if subs:
                mplayer_args = list(mplayer_args or []) + ['-sub', subs]

        resource = self._get_resource(resource)
        if resource.startswith('file://'):
            resource = resource[7:]

        self._exec('loadfile', resource, mplayer_args=mplayer_args)
        if self.volume:
            self.set_volume(volume=self.volume)

        return self.status()

    @action
    def pause(self, *_, **__):
        """Toggle the paused state"""
        self._exec('pause')
        return self.status()

    @action
    def stop(self, *_, **__):
        """Stop the playback"""
        return self.quit()

    def _cleanup(self):
        with self._cleanup_lock:
            if self._player:
                self._player.terminate()
                self._player.wait()
                try:
                    self._player.kill()
                except Exception:
                    pass

                self._player = None
                self._post_event(MediaStopEvent)

            if self._proc_monitor and os.getpid() != self._proc_monitor.pid:
                try:
                    self._proc_monitor.terminate()
                except Exception as e:
                    if self._proc_monitor or self._proc_monitor.is_alive():
                        self.logger.warning(
                            'Failed to terminate MPlayer monitor process: %s', e
                        )

                if self._proc_monitor:
                    try:
                        self._proc_monitor.join(timeout=5)
                    except AssertionError:  # Can only join a child process
                        pass

                try:
                    self._proc_monitor.kill()
                except Exception:
                    pass

                self._proc_monitor = None

    @action
    def quit(self, *_, **__):
        """Quit the player"""
        self._exec('quit')
        self._cleanup()
        return self.status()

    @action
    def voldown(self, *_, step=10.0, **__):
        """Volume down by (default: 10)%"""
        volume = (self._get_property('volume') or {}).get('volume')
        if volume is None:
            return self.status()

        return self.set_volume(volume=volume - step)

    @action
    def volup(self, *_, step=10.0, **__):
        """Volume up by (default: 10)%"""
        volume = (self._get_property('volume') or {}).get('volume')
        if volume is None:
            return self.status()

        return self.set_volume(volume=volume + step)

    @action
    def back(self, *_, offset=30.0, **__):
        """Back by (default: 30) seconds"""
        self.step_property('time_pos', -offset)
        return self.status()

    @action
    def forward(self, *_, offset=30.0, **__):
        """Forward by (default: 30) seconds"""
        self.step_property('time_pos', offset)
        return self.status()

    @action
    def toggle_subtitles(self, *_, **__):
        """Toggle the subtitles visibility"""
        response: dict = self._get_property('sub_visibility') or {}
        subs = response.get('sub_visibility')
        self._exec('sub_visibility', int(not subs))
        return self.status()

    @action
    def add_subtitles(self, filename: str, **__):
        """
        Sets media subtitles from filename

        :param filename: Subtitles file.
        """
        self._exec('sub_visibility', 1)
        self._exec('sub_load', filename)
        return self.status()

    @action
    def remove_subtitles(self, *_, index: Optional[int] = None, **__):
        """
        Removes the subtitle specified by the index (default: all)

        :param index: (1-based) index of the subtitles track to remove.
        """
        if index is None:
            self._exec('sub_remove', prefix='pausing_keep_force')
        else:
            self._exec('sub_remove', index, prefix='pausing_keep_force')

        return self.status()

    @action
    def is_playing(self, *_, **__):
        """
        :returns: True if it's playing, False otherwise
        """
        response: dict = self.get_property('pause').output or {}  # type: ignore
        return response.get('pause') is False

    @action
    def load(self, resource, *_, mplayer_args: Optional[Collection[str]] = None, **__):
        """
        Load a resource/video in the player.
        """
        if mplayer_args is None:
            mplayer_args = {}
        return self.play(resource, mplayer_args=mplayer_args)

    @action
    def mute(self, *_, **__):
        """Toggle mute state"""
        self._exec('mute', prefix='pausing_keep_force')
        return self.status()

    @action
    def seek(self, position: float, *_, **__):
        """
        Alias for :meth:`.set_position`

        :param position: Number of seconds relative to the current cursor
        """
        return self.set_position(position)

    @action
    def set_position(self, position: float, *_, **__):
        """
        Set the playback position.

        :param position: Number of seconds from the start
        """
        # cur_pos = (self._get_property('time_pos') or {}).get('time_pos')
        # if cur_pos is None:
        #     return self.status()

        # self.set_property('time_pos', position - cur_pos)
        self.set_property('time_pos', position)
        return self.status()

    @action
    def set_volume(self, volume: float, *_, **__):
        """
        Set the volume

        :param volume: Volume value between 0 and 100
        """
        self._exec('volume', max(0, min(100, volume)), 1, prefix='pausing_keep_force')
        return self.status()

    @action
    def status(self):
        """
        Get the current player state.

        :returns: A dictionary containing the current state.

        Example::

            .. code-block:: javascript

                {
                    "duration": 300.0,  // in seconds
                    "filename": "video.mp4",
                    "fullscreen": false,
                    "mute": false,
                    "name": "video.mp4",
                    "path": "/path/to/video.mp4",
                    "pause": false,
                    "percent_pos": 30.0,
                    "position": 90.0,  // in seconds
                    "seekable": true,
                    "state": "play",  // or "stop" or "pause"
                    "title": "My Video",
                    "volume": 50.0,
                    "volume_max": 100.0,
                    "url": "file:///path/to/video.mp4",
                }

        """

        if not self._player:
            return {'state': PlayerState.STOP.value}

        status = {}
        props = {
            'duration': 'length',
            'filename': 'filename',
            'fullscreen': 'fullscreen',
            'mute': 'mute',
            'name': 'filename',
            'path': 'path',
            'pause': 'pause',
            'percent_pos': 'percent_pos',
            'position': 'time_pos',
            'title': 'filename',
            'volume': 'volume',
        }

        with self._status_lock:
            for prop, player_prop in props.items():
                value = self.get_property(player_prop).output
                if isinstance(value, dict):
                    status[prop] = value.get(player_prop)

        status['seekable'] = bool(status['duration'])
        status['state'] = (
            PlayerState.PAUSE.value if status['pause'] else PlayerState.PLAY.value
        )

        if status['path']:
            status['url'] = (
                'file://' if os.path.isfile(status['path']) else ''
            ) + status['path']

        status['volume_max'] = 100

        if self._latest_resource:
            status.update(
                {
                    k: v
                    for k, v in asdict(self._latest_resource).items()
                    if v is not None
                }
            )

        return status

    def _get_property(
        self,
        property: str,  # pylint: disable=redefined-builtin
        args: Optional[Collection[str]] = None,
    ) -> dict:
        args = args or []
        response = {}
        errors = []

        result = (
            self._exec(
                'get_property',
                property,
                prefix='pausing_keep_force',
                wait_for_response=True,
                *args,
            )
            or {}
        )

        if not result:
            return response

        for k, v in result.items():
            if k == 'ERROR' and v not in errors:
                self._handle_property_error(property, args, v, errors)
            else:
                response[k] = v

        assert not errors, f'get_property errors: {errors}'
        return response

    def _handle_property_error(
        self,
        property: str,  # pylint: disable=redefined-builtin
        args: Optional[Collection[str]],
        error: str,
        errors: List[str],
    ):
        if error == 'PROPERTY_UNAVAILABLE':
            # This is a workaround to detect the end-of-file event.
            # When get_property('filename') returns PROPERTY_UNAVAILABLE
            # it means that the player is no longer playing anything
            if property == 'filename' and self._status.state != PlayerState.STOP:
                self.quit()
        else:
            errors.append(f'{property}{args}: {error}')

    @action
    def get_property(
        self,
        property: str,  # pylint: disable=redefined-builtin
        args: Optional[Collection[str]] = None,
    ):
        """
        Get a player property (e.g. pause, fullscreen etc.). See
        https://www.mplayerhq.hu/DOCS/tech/slave.txt for a full list of the
        available properties
        """
        return self._get_property(property, args=args)

    @action
    def set_property(
        self,
        property: str,  # pylint: disable=redefined-builtin
        value: Any,
        args: Optional[Collection[str]] = None,
    ):
        """
        Set a player property (e.g. pause, fullscreen etc.). See
        https://www.mplayerhq.hu/DOCS/tech/slave.txt for a full list of the
        available properties
        """

        args = args or []
        response = Response(output={})

        result = (
            self._exec(
                'set_property',
                property,
                value,
                prefix='pausing_keep_force' if property != 'pause' else None,
                wait_for_response=True,
                *args,
            )
            or {}
        )

        for k, v in result.items():
            if k == 'ERROR' and v not in response.errors:
                if not isinstance(response.errors, list):
                    response.errors = []
                response.errors.append(f'{property} {value}{args}: {v}')
            else:
                if not isinstance(response.output, dict):
                    response.output = {}
                response.output[k] = v

        return response

    @action
    def step_property(
        self,
        property: str,  # pylint: disable=redefined-builtin
        value: Any,
        *_,
        args: Optional[Collection[str]] = None,
        **__,
    ):
        """
        Step a player property (e.g. volume, time_pos etc.). See
        https://www.mplayerhq.hu/DOCS/tech/slave.txt for a full list of the
        available steppable properties
        """

        args = args or []
        response = Response(output={})

        result = (
            self._exec(
                'step_property',
                property,
                value,
                prefix='pausing_keep_force',
                wait_for_response=True,
                *args,
            )
            or {}
        )

        for k, v in result.items():
            if k == 'ERROR' and v not in response.errors:
                if not isinstance(response.errors, list):
                    response.errors = []
                response.errors.append(f'{property} {value}{args}: {v}')
            else:
                if not isinstance(response.output, dict):
                    response.output = {}
                response.output[k] = v

        return response

    def set_subtitles(self, filename: str, *_, **__):
        self.logger.debug('set_subtitles called with filename=%s', filename)
        raise NotImplementedError


# vim:sw=4:ts=4:et:



from dataclasses import dataclass, field
from typing import Optional


@dataclass
class Temperature:
    """
    System temperature sensor wrapper.
    """

    id: str = field(
        metadata={
            'metadata': {
                'description': 'Unique ID for the sensor',
                'example': 'acpi_1',
            }
        }
    )

    label: str = field(
        metadata={
            'metadata': {
                'description': 'Name of the sensor',
                'example': 'CPU',
            }
        }
    )

    value: float = field(
        metadata={
            'metadata': {
                'description': 'Current temperature value, in Celsius',
                'example': 55,
            }
        }
    )

    high: Optional[float] = field(
        metadata={
            'metadata': {
                'description': 'High threshold for the temperature sensor, in Celsius',
                'example': 75,
            }
        }
    )

    critical: Optional[float] = field(
        metadata={
            'metadata': {
                'description': 'Critical threshold for the temperature sensor, in Celsius',
                'example': 95,
            }
        }
    )



from abc import ABC, abstractmethod
from logging import getLogger
from typing import Optional, Union

import aioxmpp


# pylint: disable=too-few-public-methods
class XmppBaseMixin(ABC):
    """
    Base mixin for XMPP classes, containing common methods and properties.
    """

    DEFAULT_TIMEOUT = 20
    """Default timeout for async calls."""

    @abstractmethod
    def __init__(
        self,
        *_,
        user_id: Union[str, aioxmpp.JID],
        language: Optional[Union[str, aioxmpp.structs.LanguageTag]] = None,
        client: Optional[aioxmpp.Client] = None,
        **__,
    ):
        """
        :param user_id: Jabber/user ID, in the format ``user@example.org``.
        :param language: ISO string for the language code that will be used by
            the bot (default: ``None``).
        :param client: The main XMPP client.
        """
        self._jid = (
            aioxmpp.JID.fromstr(user_id) if isinstance(user_id, str) else user_id
        )
        """The client's registered JID."""
        self._lang = (
            aioxmpp.structs.LanguageTag.fromstr(language)
            if language and isinstance(language, str)
            else language
        )
        """The client's default language."""
        self._client: Optional[aioxmpp.Client] = client
        """The main XMPP client."""
        self.logger = getLogger(f'platypush:xmpp:{self.__class__.__name__}')

    @staticmethod
    def _jid_to_str(jid: aioxmpp.JID) -> str:
        """Convert a JID to a simple string in the format ``localpart@domain``."""
        return f'{jid.localpart}@{jid.domain}'



from collections import defaultdict
from dataclasses import dataclass, field
import logging
import threading
import time

from queue import Queue, Empty
from typing import Callable, Dict, Iterable, Type

from platypush.message import Message
from platypush.message.event import Event

logger = logging.getLogger('platypush:bus')


@dataclass
class MessageHandler:
    """
    Wrapper for a message callback handler.
    """

    msg_type: Type[Message]
    callback: Callable[[Message], None]
    kwargs: dict = field(default_factory=dict)

    def match(self, msg: Message) -> bool:
        return isinstance(msg, self.msg_type) and all(
            getattr(msg, k, None) == v for k, v in self.kwargs.items()
        )


class Bus:
    """
    Main local bus where the daemon will listen for new messages.
    """

    _MSG_EXPIRY_TIMEOUT = 60.0  # Consider a message on the bus as expired after one minute without being picked up

    def __init__(self, on_message=None):
        self.bus = Queue()
        self.on_message = on_message
        self.thread_id = threading.get_ident()
        self.handlers: Dict[
            Type[Message], Dict[Callable[[Message], None], MessageHandler]
        ] = defaultdict(dict)

        self._should_stop = threading.Event()

    def post(self, msg):
        """Sends a message to the bus"""
        self.bus.put(msg)

    def get(self):
        """Reads one message from the bus"""
        try:
            return self.bus.get(timeout=0.1)
        except Empty:
            return None

    def stop(self):
        self._should_stop.set()

    def _get_matching_handlers(
        self, msg: Message
    ) -> Iterable[Callable[[Message], None]]:
        return [
            hndl.callback
            for cls in type(msg).__mro__
            for hndl in self.handlers.get(cls, [])
            if hndl.match(msg)
        ]

    def _msg_executor(self, msg):
        def event_handler(event: Event, handler: Callable[[Event], None]):
            logger.info('Triggering event handler %s', handler.__name__)
            handler(event)

        def executor():
            for hndl in self._get_matching_handlers(msg):
                threading.Thread(target=event_handler, args=(msg, hndl)).start()

            try:
                if self.on_message:
                    self.on_message(msg)
            except Exception as e:
                logger.error('Error on processing message %s', msg)
                logger.exception(e)

        return executor

    def should_stop(self):
        return self._should_stop.is_set()

    def poll(self):
        """
        Reads messages from the bus until either stop event message or KeyboardInterrupt
        """
        if not self.on_message:
            logger.warning('No message handlers installed, cannot poll')
            return

        while not self.should_stop():
            msg = self.get()
            if msg is None:
                continue

            timestamp = (
                msg.timestamp if hasattr(msg, 'timestamp') else msg.get('timestamp')
            )
            if timestamp and time.time() - timestamp > self._MSG_EXPIRY_TIMEOUT:
                logger.debug(
                    '%f seconds old message on the bus expired, ignoring it: %s',
                    time.time() - msg.timestamp,
                    msg,
                )
                continue

            threading.Thread(target=self._msg_executor(msg)).start()

        logger.info('Bus service stopped')

    def register_handler(
        self, type: Type[Message], handler: Callable[[Message], None], **kwargs
    ) -> Callable[[], None]:
        """
        Register a generic handler to the bus.

        :param type: Type of the message to subscribe to (event inheritance also works).
        :param handler: Event handler - a function that takes a Message object as parameter.
        :param kwargs: Extra filter on the message values.
        :return: A function that can be called to remove the handler (no parameters required).
        """
        self.handlers[type][handler] = MessageHandler(type, handler, kwargs)

        def unregister():
            self.unregister_handler(type, handler)

        return unregister

    def unregister_handler(
        self, type: Type[Message], handler: Callable[[Message], None]
    ) -> None:
        """
        Remove an event handler.

        :param event_type: Event type.
        :param handler: Existing event handler.
        """
        if type not in self.handlers:
            return

        self.handlers[type].pop(handler, None)
        if len(self.handlers[type]) == 0:
            del self.handlers[type]


# vim:sw=4:ts=4:et:



from dataclasses import field
from datetime import date, datetime
from uuid import UUID

from marshmallow import (
    EXCLUDE,
    Schema,
    fields,
    post_dump,
    pre_load,
)
from marshmallow.validate import Range

from .. import Date, DateTime


def percent_field(**kwargs):
    """
    Field used to model percentage float fields between 0 and 1.
    """
    return field(
        default_factory=float,
        metadata={
            'validate': Range(min=0, max=1),
            **kwargs,
        },
    )


class DataClassSchema(Schema):
    """
    Base schema class for data classes that support Marshmallow schemas.
    """

    TYPE_MAPPING = {
        date: Date,
        datetime: DateTime,
        UUID: fields.UUID,
    }

    # pylint: disable=too-few-public-methods
    class Meta:
        """
        Exclude unknown fields.
        """

        unknown = EXCLUDE

    def _get_field(self, key: str) -> fields.Field:
        """
        Returns the matching field by either name or data_key.
        """
        if key in self.fields:
            return self.fields[key]

        matching_fields = [f for f in self.fields.values() if key == f.data_key]

        assert (
            len(matching_fields) == 1
        ), f'Could not find field {key} in {self.__class__.__name__}'

        return matching_fields[0]

    @pre_load
    def pre_load(self, data, **__) -> dict:
        return data

    @post_dump
    def post_dump(self, data: dict, **__) -> dict:
        # Use data_key parameters only for load
        new_data = {}
        for key, value in data.items():
            field = self._get_field(key)
            new_data[field.name if field.data_key is not None else key] = value

        return new_data



from typing import Iterable, List, Optional

from bluetooth_numbers import oui

from platypush.entities.bluetooth import BluetoothDevice, BluetoothService

from platypush.plugins.bluetooth.model import (
    MajorServiceClass,
    MajorDeviceClass,
    MinorDeviceClass,
)


# pylint: disable=too-few-public-methods
class BluetoothDeviceBuilder:
    """
    :class:`platypush.entity.bluetooth.BluetoothDevice` entity builder from the
    raw pybluez data.
    """

    @classmethod
    def build(
        cls,
        address: str,
        name: str,
        raw_class: int,
        services: Optional[Iterable[BluetoothService]] = None,
    ) -> BluetoothDevice:
        """
        Builds a :class:`platypush.entity.bluetooth.BluetoothDevice` from the
        raw pybluez data.
        """
        return BluetoothDevice(
            id=address,
            address=address,
            name=name,
            major_service_classes=cls._parse_major_service_classes(raw_class),
            major_device_class=cls._parse_major_device_class(raw_class),
            minor_device_classes=cls._parse_minor_device_classes(raw_class),
            manufacturer=cls._parse_manufacturer(address),
            supports_legacy=True,
            supports_ble=False,
            children=services,
        )

    @staticmethod
    def _parse_major_device_class(raw_class: int) -> MajorDeviceClass:
        """
        Parse the device major class from the raw exposed class value.
        """
        device_classes = [
            cls for cls in MajorDeviceClass if cls.value.matches(raw_class)
        ]

        return device_classes[0] if device_classes else MajorDeviceClass.UNKNOWN

    @staticmethod
    def _parse_minor_device_classes(raw_class: int) -> List[MinorDeviceClass]:
        """
        Parse the device minor classes from the raw exposed class value.
        """
        return [cls for cls in MinorDeviceClass if cls.value.matches(raw_class)]

    @staticmethod
    def _parse_major_service_classes(raw_class: int) -> List[MajorServiceClass]:
        """
        Parse the device major service classes from the raw exposed class value.
        """
        return [cls for cls in MajorServiceClass if cls.value.matches(raw_class)]

    @staticmethod
    def _parse_manufacturer(address: str) -> Optional[str]:
        """
        Parse the device manufacturer name from the raw MAC address.
        """
        return oui.get(':'.join(address.split(':')[:3]).upper())



import os
from dataclasses import dataclass
from datetime import datetime
from enum import IntEnum
from typing import Optional, List, Union, Dict

from platypush.context import Variable
from platypush.message.event.nextcloud import NextCloudActivityEvent
from platypush.plugins import RunnablePlugin, action


@dataclass
class ClientConfig:
    """
    Configuration for the NextCloud client.
    """

    url: str
    username: str
    password: str

    def to_dict(self):
        return {
            'url': self.url,
            'username': self.username,
            'password': self.password,
        }


class ShareType(IntEnum):
    """
    Enumerates the types of shares that can be created.
    """

    USER = 0
    GROUP = 1
    PUBLIC_LINK = 3
    EMAIL = 4
    FEDERATED_CLOUD_SHARE = 6
    CIRCLE = 7
    TALK_CONVERSATION = 10


class Permission(IntEnum):
    """
    Enumerates the permissions that can be granted to a user/group on a
    resource.
    """

    READ = 1
    UPDATE = 2
    CREATE = 4
    DELETE = 8
    SHARE = 16
    ALL = 31


_LAST_ACTIVITY_VARNAME = '_NEXTCLOUD_LAST_ACTIVITY_ID'


class NextcloudPlugin(RunnablePlugin):
    """
    Plugin to interact with a NextCloud instance.
    """

    def __init__(
        self,
        url: str,
        username: str,
        password: str,
        poll_interval: Optional[float] = 30.0,
        **kwargs,
    ):
        """
        :param url: URL to the index of your default NextCloud instance.
        :param username: Default NextCloud username.
        :param password: Default NextCloud password. If the user has 2FA
            enabled then you can generate an app password from the personal
            settings page of your NextCloud instance.
        :param poll_interval: How often the plugin should poll for new activity
            events (default: 30 seconds).
        """
        super().__init__(poll_interval=poll_interval, **kwargs)
        self.conf = ClientConfig(url=url, username=username, password=password)
        self._client = self._get_client(**self.conf.to_dict(), raise_on_empty=False)
        self._last_seen_id = None

    def _get_client(
        self,
        url: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        raise_on_empty: bool = False,
    ):
        from nextcloud import NextCloud

        if not url:
            if not self.conf.url:
                if raise_on_empty:
                    raise AssertionError('No url/username/password provided')
                return None

            return NextCloud(
                endpoint=self.conf.url,
                user=self.conf.username,
                password=self.conf.password,
            )

        return NextCloud(endpoint=url, user=username, password=password)

    @property
    def _last_seen_var(self) -> Variable:
        return Variable(_LAST_ACTIVITY_VARNAME)

    @property
    def last_seen_id(self) -> int:
        if self._last_seen_id is None:
            self._last_seen_id = self._last_seen_var.get() or 0

        return int(self._last_seen_id)

    @last_seen_id.setter
    def last_seen_id(self, value: int = 0):
        self._last_seen_var.set(value)
        self._last_seen_id = value

    @staticmethod
    def _activity_to_event(activity: dict) -> NextCloudActivityEvent:
        activity_type = activity.pop('type')
        dt = datetime.fromisoformat(activity.pop('datetime'))
        return NextCloudActivityEvent(
            activity_type=activity_type, datetime=dt, **activity
        )

    @staticmethod
    def _get_permissions(permissions: Optional[List[str]]) -> int:
        int_perm = 0

        for perm in permissions or []:
            perm = perm.upper()
            assert hasattr(Permission, perm), (
                f'Unknown permissions type: {perm}. Supported permissions: '
                f'{[p.name.lower() for p in Permission]}'
            )

            if perm == 'ALL':
                int_perm = Permission.ALL.value
                break

            int_perm += getattr(Permission, perm).value

        return int_perm

    @staticmethod
    def _get_share_type(share_type: str) -> int:
        share_type = share_type.upper()
        assert hasattr(ShareType, share_type), (
            f'Unknown share type: {share_type}. Supported share types: '
            f'{[s.name.lower() for s in ShareType]}'
        )

        return getattr(ShareType, share_type).value

    def _execute(self, server_args: dict, method: str, *args, **kwargs):
        client = self._get_client(**server_args)
        func = getattr(client, method, None)
        assert func, f'No such NextCloud method: {method}'

        response = func(*args, **kwargs)
        assert response is not None, 'No response from NextCloud server'
        assert response.is_ok, (
            'Error on '
            + method
            + '('
            + ', '.join(args)
            + (', ' if args and kwargs else '')
            + ', '.join([f'{k}={v}' for k, v in kwargs.items()])
            + '): status='
            + str(response.status_code)
            + ', message='
            + getattr(response, 'meta', {}).get('message', '[No message]')
            + ', data='
            + str(response.json_data)
        )

        return response.json_data

    def _get_activities(
        self,
        since: Optional[str] = None,
        limit: Optional[int] = 25,
        object_type: Optional[str] = None,
        object_id: Optional[int] = None,
        sort: str = 'desc',
        **server_args,
    ) -> List[dict]:
        return self._execute(
            server_args,
            'get_activities',
            since=since,
            limit=limit,
            object_type=object_type,
            object_id=object_id,
            sort=sort,
        )

    @action
    def get_activities(
        self,
        since: Optional[str] = None,
        limit: Optional[int] = 25,
        object_type: Optional[str] = None,
        object_id: Optional[int] = None,
        sort: str = 'desc',
        **server_args,
    ) -> List[dict]:
        """
        Get the list of recent activities on an instance.

        :param since: Only return the activities that have occurred since the
            specified ID.
        :param limit: Maximum number of activities to be returned (default: 25).
        :param object_type: Filter by object type.
        :param object_id: Only get the activities related to a specific
            ``object_id``.
        :param sort: Sort mode, ``asc`` for ascending, ``desc`` for descending
            (default: ``desc``).
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        :return: The list of selected activities.
        """
        return self._get_activities(
            since=since,
            limit=limit,
            object_type=object_type,
            object_id=object_id,
            sort=sort,
            **server_args,
        )

    @action
    def get_apps(self, **server_args) -> List[str]:
        """
        Get the list of apps installed on a NextCloud instance.

        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        :return: The list of installed apps as strings.
        """
        return self._execute(server_args, 'get_apps').get('apps', [])

    @action
    def enable_app(self, app_id: Union[str, int], **server_args):
        """
        Enable an app.

        :param app_id: App ID.
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'enable_app', app_id)

    @action
    def disable_app(self, app_id: Union[str, int], **server_args):
        """
        Disable an app.

        :param app_id: App ID.
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'disable_app', app_id)

    @action
    def get_app(self, app_id: Union[str, int], **server_args) -> dict:
        """
        Provides information about an application.

        :param app_id: App ID.
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        return self._execute(server_args, 'get_app', app_id)

    @action
    def get_capabilities(self, **server_args) -> dict:
        """
        Returns the capabilities of the server.

        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        :return: Server capabilities. Example:

          .. code-block:: json

            {
              "version": {
                "major": 27,
                "minor": 1,
                "micro": 5,
                "string": "27.1.5",
                "edition": "",
                "extendedSupport": false
              },
              "capabilities": {
                "core": {
                  "pollinterval": 60,
                  "webdav-root": "remote.php/webdav",
                },
                "files": {
                  "bigfilechunking": true,
                  "blacklisted_files": [
                    ".htaccess"
                  ],
                  "comments": true,
                  "undelete": true,
                  "versioning": true,
                  "version_labeling": true,
                  "version_deletion": true
                },
                "activity": {
                  "apiv2": [
                    "filters",
                    "filters-api",
                    "previews",
                    "rich-strings"
                  ]
                },
                "notifications": {
                  "ocs-endpoints": [
                    "list",
                    "get",
                    "delete",
                    "delete-all",
                    "icons",
                    "rich-strings",
                    "action-web",
                    "user-status",
                    "exists"
                  ],
                  "push": [
                    "devices",
                    "object-data",
                    "delete"
                  ],
                  "admin-notifications": [
                    "ocs",
                    "cli"
                  ]
                },
                "theming": {
                  "name": "Nextcloud",
                  "url": "https://nextcloud.com",
                  "slogan": "a safe home for all your data",
                  "color": "#1F5C98",
                  "color-text": "#ffffff",
                  "color-element": "#1F5C98",
                  "color-element-bright": "#1F5C98",
                  "color-element-dark": "#1F5C98",
                  "logo": "https://cloud.example.com/core/img/logo/logo.svg?v=0",
                  "background": "#1F5C98",
                  "background-plain": true,
                  "background-default": true,
                  "logoheader": "https://cloud.example.com/core/img/logo/logo.svg?v=0",
                  "favicon": "https://cloud.example.com/core/img/logo/logo.svg?v=0"
                },
                "user_status": {
                  "enabled": true,
                  "restore": true,
                  "supports_emoji": true
                },
              }
            }

        """
        return self._execute(server_args, 'get_capabilities')

    @action
    def add_group(self, group_id: Union[str, int], **server_args):
        """
        Create a new group.

        :param group_id: New group unique ID.
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'add_group', group_id)

    @action
    def delete_group(self, group_id: Union[str, int], **server_args):
        """
        Delete a group.

        :param group_id: Group ID.
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'delete_group', group_id)

    @action
    def get_group(self, group_id: Union[str, int], **server_args) -> dict:
        """
        Get the information of a group.

        :param group_id: Group ID.
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        return self._execute(server_args, 'get_group', group_id)

    @action
    def get_groups(
        self,
        search: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        **server_args,
    ) -> List[str]:
        """
        Search for groups.

        :param search: Search for groups matching the specified substring.
        :param limit: Maximum number of returned entries.
        :param offset: Start offset.
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        return self._execute(
            server_args, 'get_groups', search=search, limit=limit, offset=offset
        ).get('groups', [])

    @action
    def create_group_folder(self, name: str, **server_args):
        """
        Create a new group folder.

        :param name: Name/path of the folder.
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'create_group_folder', name)

    @action
    def delete_group_folder(self, folder_id: Union[int, str], **server_args):
        """
        Delete a new group folder.

        :param folder_id: Folder ID.
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'delete_group_folder', folder_id)

    @action
    def get_group_folder(self, folder_id: Union[int, str], **server_args) -> dict:
        """
        Get a new group folder.

        :param folder_id: Folder ID.
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        return self._execute(server_args, 'get_group_folder', folder_id)

    @action
    def get_group_folders(self, **server_args) -> List:
        """
        Get the list new group folder.

        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        return self._execute(server_args, 'get_group_folders')

    @action
    def rename_group_folder(
        self, folder_id: Union[int, str], new_name: str, **server_args
    ):
        """
        Rename a group folder.

        :param folder_id: Folder ID.
        :param new_name: New folder name.
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'rename_group_folder', folder_id, new_name)

    @action
    def grant_access_to_group_folder(
        self, folder_id: Union[int, str], group_id: str, **server_args
    ):
        """
        Grant access to a group folder to a given group.

        :param folder_id: Folder ID.
        :param group_id: Group ID.
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'grant_access_to_group_folder', folder_id, group_id)

    @action
    def revoke_access_to_group_folder(
        self, folder_id: Union[int, str], group_id: str, **server_args
    ):
        """
        Revoke access to a group folder to a given group.

        :param folder_id: Folder ID.
        :param group_id: Group ID.
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'revoke_access_to_group_folder', folder_id, group_id)

    @action
    def set_group_folder_quota(
        self, folder_id: Union[int, str], quota: Optional[int], **server_args
    ):
        """
        Set the quota of a group folder.

        :param folder_id: Folder ID.
        :param quota: Quota in bytes - set None for unlimited.
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        self._execute(
            server_args,
            'set_quota_of_group_folder',
            folder_id,
            quota if quota is not None else -3,
        )

    @action
    def set_group_folder_permissions(
        self,
        folder_id: Union[int, str],
        group_id: str,
        permissions: List[str],
        **server_args,
    ):
        """
        Set the permissions on a folder for a group.

        :param folder_id: Folder ID.
        :param group_id: Group ID.
        :param permissions: New permissions, as a list including any of the
            following:

              - ``read``
              - ``update``
              - ``create``
              - ``delete``
              - ``share``
              - ``all``

        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        self._execute(
            server_args,
            'set_permissions_to_group_folder',
            folder_id,
            group_id,
            self._get_permissions(permissions),
        )

    @action
    def get_notifications(self, **server_args) -> List[dict]:
        """
        Get the list of notifications for the logged user.

        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        :return: A list of notifications. Example:

          .. code-block:: json

            [
              {
                "notification_id": 1234,
                "app": "updatenotification",
                "user": "username",
                "datetime": "2024-01-01T19:00:00+00:00",
                "object_type": "side_menu",
                "object_id": "1.2.3",
                "subject": "Update for MyApp to version 1.2.3 is available.",
                "message": "",
                "link": "https://cloud.example.com/index.php/settings/apps/updates#myapp",
                "subjectRich": "Update for {app} to version 1.2.3 is available.",
                "subjectRichParameters": {
                  "app": {
                    "type": "app",
                    "id": "myapp",
                    "name": "MyApp"
                  }
                },
                "messageRich": "",
                "messageRichParameters": [],
                "icon": "https://cloud.example.com/apps/updatenotification/img/notification.svg",
                "shouldNotify": true,
                "actions": []
              }
            ]

        """
        return self._execute(server_args, 'get_notifications')

    @action
    def delete_notifications(self, **server_args):
        """
        Delete all notifications for the logged user.

        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'delete_all_notifications')

    @action
    def get_notification(self, notification_id: int, **server_args) -> dict:
        """
        Get the content of a notification.

        :param notification_id: Notification ID.
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        :return: Notification details. Example:

          .. code-block:: json

              {
                "notification_id": 1234,
                "app": "updatenotification",
                "user": "username",
                "datetime": "2024-01-01T19:00:00+00:00",
                "object_type": "side_menu",
                "object_id": "1.2.3",
                "subject": "Update for MyApp to version 1.2.3 is available.",
                "message": "",
                "link": "https://cloud.example.com/index.php/settings/apps/updates#myapp",
                "subjectRich": "Update for {app} to version 1.2.3 is available.",
                "subjectRichParameters": {
                  "app": {
                    "type": "app",
                    "id": "myapp",
                    "name": "MyApp"
                  }
                },
                "messageRich": "",
                "messageRichParameters": [],
                "icon": "https://cloud.example.com/apps/updatenotification/img/notification.svg",
                "shouldNotify": true,
                "actions": []
              }

        """
        return self._execute(server_args, 'get_notification', notification_id)

    @action
    def delete_notification(self, notification_id: int, **server_args):
        """
        Delete a notification.

        :param notification_id: Notification ID.
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'delete_notification', notification_id)

    @action
    def create_share(
        self,
        path: str,
        share_type: str,
        share_with: Optional[str] = None,
        public_upload: bool = False,
        password: Optional[str] = None,
        permissions: Optional[List[str]] = None,
        **server_args,
    ) -> dict:
        """
        Share a file/folder with a user/group or a public link.

        :param path: Path to the resource to be shared.
        :param share_type: Share type. Supported values:

          - ``user``
          - ``group``
          - ``public_link``
          - ``email``
          - ``federated_cloud_share``
          - ``circle``
          - ``talk_conversation``

        :param share_with: User/group ID, email or conversation ID the resource
            should be shared with.
        :param public_upload: Whether public upload to the shared folder is
            allowed (default: False).
        :param password: Optional password to protect the share.
        :param permissions: Share permissions, as a list including any of the
            following (default: ``read``):

              - ``read``
              - ``update``
              - ``create``
              - ``delete``
              - ``share``
              - ``all``

        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        :return: The details of the newly created share. Example:

          .. code-block:: json

            {
              "id": "4",
              "share_type": 3,
              "uid_owner": "your_uid",
              "displayname_owner": "Your Name",
              "permissions": 17,
              "can_edit": true,
              "can_delete": true,
              "stime": 1599691325,
              "parent": null,
              "expiration": null,
              "token": "AbCdEfG0123456789",
              "uid_file_owner": "your_uid",
              "note": "",
              "label": "",
              "displayname_file_owner": "Your Name",
              "path": "/Shared Path",
              "item_type": "folder",
              "mimetype": "httpd/unix-directory",
              "storage_id": "home::your-uid",
              "storage": 2,
              "item_source": 13960,
              "file_source": 13960,
              "file_parent": 6,
              "file_target": "/Shared Path",
              "share_with": null,
              "share_with_displayname": "(Shared link)",
              "password": null,
              "send_password_by_talk": false,
              "url": "https://your-domain/nextcloud/index.php/s/AbCdEfG0123456789",
              "mail_send": 1,
              "hide_download": 0
            }

        """
        share_type_id = self._get_share_type(share_type)
        permissions_id = self._get_permissions(permissions or ['read'])
        return self._execute(
            server_args,
            'create_share',
            path,
            share_type=share_type_id,
            share_with=share_with,
            public_upload=public_upload,
            password=password,
            permissions=permissions_id,
        )

    @action
    def get_shares(self, **server_args) -> List[dict]:
        """
        Get the list of shares available on the server.

        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        :return: List of available shares. Example:

          .. code-block:: json

            [
                {
                  "id": "4",
                  "share_type": 3,
                  "uid_owner": "your_uid",
                  "displayname_owner": "Your Name",
                  "permissions": 17,
                  "can_edit": true,
                  "can_delete": true,
                  "stime": 1599691325,
                  "parent": null,
                  "expiration": null,
                  "token": "AbCdEfG0123456789",
                  "uid_file_owner": "your_uid",
                  "note": "",
                  "label": "",
                  "displayname_file_owner": "Your Name",
                  "path": "/Shared Path",
                  "item_type": "folder",
                  "mimetype": "httpd/unix-directory",
                  "storage_id": "home::your-uid",
                  "storage": 2,
                  "item_source": 13960,
                  "file_source": 13960,
                  "file_parent": 6,
                  "file_target": "/Shared Path",
                  "share_with": null,
                  "share_with_displayname": "(Shared link)",
                  "password": null,
                  "send_password_by_talk": false,
                  "url": "https://your-domain/nextcloud/index.php/s/AbCdEfG0123456789",
                  "mail_send": 1,
                  "hide_download": 0
                }
            ]

        """
        return self._execute(server_args, 'get_shares')

    @action
    def delete_share(self, share_id: int, **server_args):
        """
        Remove the shared state of a resource.

        :param share_id: Share ID.
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'delete_share', str(share_id))

    @action
    def get_share(self, share_id: int, **server_args) -> dict:
        """
        Get the information of a shared resource.

        :param share_id: Share ID.
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        :return: The details of the share. Example:

          .. code-block:: json

              {
                "id": "4",
                "share_type": 3,
                "uid_owner": "your_uid",
                "displayname_owner": "Your Name",
                "permissions": 17,
                "can_edit": true,
                "can_delete": true,
                "stime": 1599691325,
                "parent": null,
                "expiration": null,
                "token": "AbCdEfG0123456789",
                "uid_file_owner": "your_uid",
                "note": "",
                "label": "",
                "displayname_file_owner": "Your Name",
                "path": "/Shared Path",
                "item_type": "folder",
                "mimetype": "httpd/unix-directory",
                "storage_id": "home::your-uid",
                "storage": 2,
                "item_source": 13960,
                "file_source": 13960,
                "file_parent": 6,
                "file_target": "/Shared Path",
                "share_with": null,
                "share_with_displayname": "(Shared link)",
                "password": null,
                "send_password_by_talk": false,
                "url": "https://your-domain/nextcloud/index.php/s/AbCdEfG0123456789",
                "mail_send": 1,
                "hide_download": 0
              }

        """
        return self._execute(server_args, 'get_share_info', str(share_id))

    @action
    def update_share(
        self,
        share_id: int,
        public_upload: Optional[bool] = None,
        password: Optional[str] = None,
        permissions: Optional[List[str]] = None,
        expire_date: Optional[str] = None,
        **server_args,
    ):
        """
        Update the permissions of a shared resource.

        :param share_id: Share ID.
        :param public_upload: Whether public upload to the shared folder is
            allowed (default: False).
        :param password: Optional password to protect the share.
        :param permissions: Share permissions, as a list including any of the
            following (default: ``read``):

              - ``read``
              - ``update``
              - ``create``
              - ``delete``
              - ``share``
              - ``all``

        :param expire_date: Share expiration date, in the format ``YYYY-MM-DD``.
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        perms = self._get_permissions(permissions) if permissions else None
        self._execute(
            server_args,
            'update_share',
            share_id,
            public_upload=public_upload,
            password=password,
            permissions=perms,
            expire_date=expire_date,
        )

    @action
    def create_user(self, user_id: str, password: str, **server_args):
        """
        Create a user.

        :param user_id: User ID/name.
        :param password: User password
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'add_user', user_id, password)

    @action
    def edit_user(self, user_id: str, properties: Dict[str, str], **server_args):
        """
        Update a set of properties of a user.

        :param user_id: User ID/name.
        :param properties: Key-value pair of user attributes to be edited.
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        for k, v in properties.items():
            self._execute(server_args, 'edit_user', user_id, k, v)

    @action
    def get_user(self, user_id: str, **server_args) -> dict:
        """
        Get the details of a user.

        :param user_id: User ID/name.
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        :return: User details. Example:

          .. code-block:: json

            {
              "enabled": true,
              "storageLocation": "/mnt/hd/nextcloud/user",
              "id": "user",
              "lastLogin": 1599693750000,
              "backend": "Database",
              "subadmin": [],
              "quota": {
                "free": 6869434515456,
                "used": 1836924441,
                "total": 6871271439897,
                "relative": 0.03,
                "quota": -3
              },
              "email": "info@yourdomain.com",
              "displayname": "Your Name",
              "phone": "+1234567890",
              "address": "",
              "website": "https://yourdomain.com",
              "twitter": "@You",
              "groups": [
                "admin"
              ],
              "language": "en",
              "locale": "",
              "backendCapabilities": {
                "setDisplayName": true,
                "setPassword": true
              }
            }

        """
        return self._execute(server_args, 'get_user', user_id)

    @action
    def get_users(
        self,
        search: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        **server_args,
    ) -> List[str]:
        """
        Get the list of users matching some search criteria.

        :param search: Return users matching the provided string.
        :param limit: Maximum number of results to be returned (default: no
            limit).
        :param offset: Search results offset (default: None).
        :return: List of the matched user IDs. Example:

            .. code-block:: json

                [
                    {
                      "enabled": true,
                      "storageLocation": "/mnt/hd/nextcloud/user",
                      "id": "user",
                      "lastLogin": 1599693750000,
                      "backend": "Database",
                      "subadmin": [],
                      "quota": {
                        "free": 6869434515456,
                        "used": 1836924441,
                        "total": 6871271439897,
                        "relative": 0.03,
                        "quota": -3
                      },
                      "email": "info@yourdomain.com",
                      "displayname": "Your Name",
                      "phone": "+1234567890",
                      "address": "",
                      "website": "https://yourdomain.com",
                      "twitter": "@You",
                      "groups": [
                        "admin"
                      ],
                      "language": "en",
                      "locale": "",
                      "backendCapabilities": {
                        "setDisplayName": true,
                        "setPassword": true
                      }
                    }
                ]

        """
        return self._execute(
            server_args, 'get_users', search=search, limit=limit, offset=offset
        )

    @action
    def delete_user(self, user_id: str, **server_args):
        """
        Delete a user.

        :param user_id: User ID/name.
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'delete_user', user_id)

    @action
    def enable_user(self, user_id: str, **server_args):
        """
        Enable a user.

        :param user_id: User ID/name.
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'enable_user', user_id)

    @action
    def disable_user(self, user_id: str, **server_args):
        """
        Disable a user.

        :param user_id: User ID/name.
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'disable_user', user_id)

    @action
    def add_to_group(self, user_id: str, group_id: str, **server_args):
        """
        Add a user to a group.

        :param user_id: User ID/name.
        :param group_id: Group ID.
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'add_to_group', user_id, group_id)

    @action
    def remove_from_group(self, user_id: str, group_id: str, **server_args):
        """
        Remove a user from a group.

        :param user_id: User ID/name.
        :param group_id: Group ID.
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'remove_from_group', user_id, group_id)

    @action
    def create_subadmin(self, user_id: str, group_id: str, **server_args):
        """
        Add a user as a subadmin for a group.

        :param user_id: User ID/name.
        :param group_id: Group ID.
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'create_subadmin', user_id, group_id)

    @action
    def remove_subadmin(self, user_id: str, group_id: str, **server_args):
        """
        Remove a user as a subadmin from a group.

        :param user_id: User ID/name.
        :param group_id: Group ID.
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'remove_subadmin', user_id, group_id)

    @action
    def get_subadmin_groups(self, user_id: str, **server_args) -> List[str]:
        """
        Get the groups where a given user is subadmin.

        :param user_id: User ID/name.
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        :return: List of matched groups as strings.
        """
        return self._execute(server_args, 'get_subadmin_groups', user_id)

    @action
    def create_folder(self, path: str, **server_args):
        """
        Create a folder.

        :param path: Path to the folder.
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'create_folder', path)

    @action
    def mkdir(self, path: str, **server_args):
        """
        Alias for :meth:`.create_folder`.

        :param path: Path to the folder.
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        self.create_folder(path, **server_args)

    @action
    def delete(self, path: str, **server_args):
        """
        Delete a file or folder.

        :param path: Path to the resource.
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'delete_path', path)

    @action
    def upload_file(
        self,
        remote_path: str,
        local_path: Optional[str] = None,
        content: Optional[str] = None,
        timestamp: Optional[Union[datetime, int, str]] = None,
        **server_args,
    ):
        """
        Upload a file.

        :param remote_path: Path to the remote resource.
        :param local_path: If set, identifies the path to the local file to be
            uploaded.
        :param content: If set, create a new file with this content instead of
            uploading an existing file.
        :param timestamp: File timestamp. If not specified it will be retrieved
            from the file info or set to ``now`` if ``content`` is specified.
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        if isinstance(timestamp, str):
            timestamp = datetime.fromisoformat(timestamp)
        if isinstance(timestamp, datetime):
            timestamp = int(timestamp.timestamp())

        assert (local_path or content) and not (
            local_path and content
        ), 'Please specify either local_path or content'

        if local_path:
            method = 'upload_file'
            local_path = os.path.abspath(os.path.expanduser(local_path))
        else:
            method = 'upload_file_contents'

        return self._execute(
            server_args,
            method,
            local_path or content,
            remote_path,
            timestamp=timestamp,
        )

    @action
    def download_file(
        self,
        remote_path: str,
        local_path: str,
        **server_args,
    ):
        """
        Download a file.

        :param remote_path: Path to the remote resource.
        :param local_path: Path to the local folder.
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        local_path = os.path.abspath(os.path.expanduser(local_path))
        cur_dir = os.getcwd()

        try:
            os.chdir(local_path)
            return self._execute(server_args, 'download_file', remote_path)
        finally:
            os.chdir(cur_dir)

    @action
    def list(
        self,
        path: str,
        depth: int = 1,
        all_properties: bool = False,
        **server_args,
    ) -> List[dict]:
        """
        List the content of a folder on the NextCloud instance.

        :param path: Remote path.
        :param depth: Search depth (default: 1).
        :param all_properties: Return all the file properties available
            (default: ``False``).
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        return self._execute(
            server_args,
            'list_folders',
            path,
            depth=depth,
            all_properties=all_properties,
        )

    @action
    def list_favorites(self, path: Optional[str] = None, **server_args) -> List[dict]:
        """
        List the favorite items for a user.

        :param path: Return only the favorites under this path.
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        return self._execute(server_args, 'list_folders', path)

    @action
    def mark_favorite(self, path: Optional[str] = None, **server_args):
        """
        Add a path to a user's favorites.

        :param path: Resource path.
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'set_favorites', path)

    @action
    def copy(
        self,
        path: str,
        destination: str,
        overwrite: bool = False,
        **server_args,
    ):
        """
        Copy a resource to another path.

        :param path: Resource path.
        :param destination: Destination path.
        :param overwrite: Set to ``True`` if you want to overwrite any existing
            file (default: ``False``).
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'copy_path', path, destination, overwrite=overwrite)

    @action
    def move(
        self,
        path: str,
        destination: str,
        overwrite: bool = False,
        **server_args,
    ):
        """
        Move a resource to another path.

        :param path: Resource path.
        :param destination: Destination path.
        :param overwrite: Set to ``True`` if you want to overwrite any existing
            file (default: ``False``).
        :param server_args: Override the default server settings (see
            :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'move_path', path, destination, overwrite=overwrite)

    def main(self):
        while not self.should_stop():
            last_seen_id = self.last_seen_id
            new_last_seen_id = last_seen_id
            activities = self._get_activities(
                sort='desc',
                url=self.conf.url,
                username=self.conf.username,
                password=self.conf.password,
                limit=100,
            )

            events = []
            for activity in activities:
                activity_id = int(activity['activity_id'])
                if last_seen_id and activity_id <= last_seen_id:
                    break

                events.append(self._activity_to_event(activity))

                if not new_last_seen_id or activity_id > new_last_seen_id:
                    new_last_seen_id = activity_id

            for evt in events[::-1]:
                self._bus.post(evt)

            if new_last_seen_id and last_seen_id != new_last_seen_id:
                self.last_seen_id = new_last_seen_id

            self.wait_stop(self.poll_interval)


# vim:sw=4:ts=4:et:



from abc import ABC, abstractmethod
from typing import Collection, Iterable, List, Mapping, Optional, Tuple, Type, Union

from platypush.common.sensors import Numeric, SensorDataType
from platypush.context import get_bus
from platypush.entities import Entity
from platypush.entities.managers.sensors import SensorEntityManager
from platypush.message.event.sensor import (
    SensorDataAboveThresholdEvent,
    SensorDataBelowThresholdEvent,
    SensorDataChangeEvent,
    SensorDataEvent,
)
from platypush.plugins import RunnablePlugin, action
from platypush.utils import get_plugin_name_by_class

NoneType = type(None)
ThresholdType = Union[Numeric, Tuple[Numeric, Numeric]]
ThresholdConfiguration = Union[ThresholdType, Mapping[str, ThresholdType]]


class SensorPlugin(RunnablePlugin, SensorEntityManager, ABC):
    """
    Sensor abstract plugin. Any plugin that interacts with sensors
    should implement this class.
    """

    _max_retry_secs = 60.0
    """
    In case of failure, we apply an exponential back-off retry algorithm. This
    is the maximum number of seconds that we should wait during these retries.
    """

    def __init__(
        self,
        thresholds: Optional[ThresholdConfiguration] = None,
        tolerance: SensorDataType = 0,
        enabled_sensors: Optional[Iterable[str]] = None,
        **kwargs,
    ):
        """
        :param thresholds: A number, numeric pair or mapping of ``str`` to
            number/numeric pair representing the thresholds for the sensor.

            Examples:

                .. code-block:: yaml

                    # Any value below 25 from any sensor will trigger a
                    # SensorDataBelowThresholdEvent, if the previous value was
                    # equal or above, and any value above 25 will trigger a
                    # SensorDataAboveThresholdEvent, if the previous value was
                    # equal or below
                    thresholds: 25.0

                    # Same as above, but the threshold is only applied to
                    # ``temperature`` readings
                    thresholds:
                        temperature: 25.0

                    # Any value below 20 from any sensor will trigger a
                    # SensorDataBelowThresholdEvent, if the previous value was
                    # equal or above, and any value above 25 will trigger a
                    # SensorDataAboveThresholdEvent, if the previous value was
                    # equal or below (hysteresis configuration with double
                    # threshold)
                    thresholds:
                        - 20.0
                        - 25.0

                    # Same as above, but the threshold is only applied to
                    # ``temperature`` readings
                    thresholds:
                        temperature:
                            - 20.0
                            - 25.0

        :param tolerance: If set, then the sensor change events will be
            triggered only if the difference between the new value and the
            previous value is higher than the specified tolerance. For example,
            if the sensor data is mapped to a dictionary::

                {
                    "temperature": 0.01,  # Tolerance on the 2nd decimal digit
                    "humidity": 0.1       # Tolerance on the 1st decimal digit
                }

            Or, if it's a raw scalar number::

                0.1  # Tolerance on the 1st decimal digit

            Or, if it's a list of values::

                [
                    0.01,   # Tolerance on the 2nd decimal digit for the first value
                    0.1     # Tolerance on the 1st decimal digit for the second value
                ]

        :param enabled_sensors: If :meth:`.get_measurement` returns a key-value
            mapping, and ``enabled_sensors`` is set, then only the reported
            sensor keys will be returned.
        """
        super().__init__(**kwargs)
        self._tolerance = tolerance
        self._thresholds = thresholds
        self._enabled_sensors = set(enabled_sensors or [])
        self._last_measurement: Optional[SensorDataType] = None
        """ Latest measurement from the sensor. """

    def _has_changes_scalar(
        self,
        old_data: Union[int, float],
        new_data: Union[int, float],
        attr: Optional[str] = None,
        index: Optional[int] = None,
    ) -> bool:
        """
        Returns ``True`` if the new data has changes compared to the old data -
        limited to numeric scalar values.
        """
        if isinstance(self._tolerance, (int, float)):
            tolerance = self._tolerance
        elif isinstance(self._tolerance, dict) and attr:
            tolerance = self._tolerance.get(attr, 0)  # type: ignore
        elif isinstance(self._tolerance, (list, tuple)) and index:
            tolerance = self._tolerance[index]
        else:
            tolerance = 0

        return abs(old_data - new_data) > tolerance

    def _has_changes(
        self,
        old_data: Optional[SensorDataType],
        new_data: Optional[SensorDataType],
        attr: Optional[str] = None,
        index: Optional[int] = None,
    ) -> bool:
        """
        Returns ``True`` if the new data has changes compared to the old data.
        It also applies the configured tolerance thresholds.
        """
        # If there is no previous data, then the new data will always be a change
        if old_data is None:
            return True

        # If the new data is missing, then we have no new changes
        if new_data is None:
            return False

        # If the data is scalar, then run the comparison logic
        if isinstance(old_data, (int, float)) and isinstance(new_data, (int, float)):
            return self._has_changes_scalar(old_data, new_data, attr, index)

        # If the data is dict-like, recursively call _has_changes on its attributes
        if isinstance(old_data, dict) and isinstance(new_data, dict):
            return any(
                self._has_changes(old_data.get(attr), value, attr=attr)  # type: ignore
                for attr, value in new_data.items()
            )

        # If the data is list-like, recursively call _has_changes on its values
        if isinstance(old_data, (list, tuple)) and isinstance(new_data, (list, tuple)):
            return any(
                self._has_changes(old_data[i], value, index=i)
                for i, value in enumerate(new_data)
            )

        return old_data != new_data

    def _process_scalar_threshold_events(
        self,
        old_data: Optional[Numeric],
        new_data: Numeric,
        attr: Optional[str] = None,
    ) -> List[SensorDataEvent]:
        """
        Inner scalar processing for sensor above/below threshold events.
        """
        event_types: List[Type[SensorDataEvent]] = []
        event_args = {
            'source': get_plugin_name_by_class(self.__class__),
        }

        # If we're mapping against a dict attribute, extract its thresholds,
        # otherwise use the default configured thresholds
        thresholds = (
            self._thresholds.get(attr)
            if attr and isinstance(self._thresholds, dict)
            else self._thresholds
        )

        # Normalize low/high thresholds
        low_t, high_t = (
            sorted(thresholds[:2])
            if isinstance(thresholds, (list, tuple))
            else (thresholds, thresholds)
        )

        if low_t is None or high_t is None:
            return []

        assert isinstance(low_t, (int, float)) and isinstance(
            high_t, (int, float)
        ), f'Non-numeric thresholds detected: "{low_t}" and "{high_t}"'

        # Above threshold case
        if (old_data is None or old_data <= high_t) and new_data > high_t:
            event_types.append(SensorDataAboveThresholdEvent)
        # Below threshold case
        elif (old_data is None or old_data >= low_t) and new_data < low_t:
            event_types.append(SensorDataBelowThresholdEvent)

        return [
            event_type(
                data={attr: new_data} if attr else new_data,
                **event_args,
            )
            for event_type in event_types
        ]

    def _process_threshold_events(
        self,
        old_data: Optional[SensorDataType],
        new_data: SensorDataType,
        attr: Optional[str] = None,
    ) -> List[SensorDataEvent]:
        """
        Processes sensor above/below threshold events.
        """
        events: List[SensorDataEvent] = []

        # If there are no configured thresholds, there's nothing to do
        if self._thresholds in (None, {}, (), []):
            return events

        # Scalar case
        if isinstance(old_data, (int, float, NoneType)) and isinstance(
            new_data, (int, float)
        ):
            return self._process_scalar_threshold_events(
                old_data, new_data, attr  # type: ignore
            )

        # From here on, threshold comparison only applies if both the old and
        # new data is a str -> number mapping
        if not (isinstance(old_data, (dict, NoneType)) and isinstance(new_data, dict)):
            return events

        # Recursively call _process_threshold_events on the data attributes
        for attr, value in new_data.items():  # type: ignore
            events.extend(
                self._process_threshold_events(
                    old_data=(
                        old_data.get(attr)  # type: ignore
                        if isinstance(old_data, dict)
                        else old_data
                    ),
                    new_data=value,
                    attr=str(attr),
                )
            )

        return events

    def _process_sensor_events(
        self, old_data: Optional[SensorDataType], new_data: Optional[SensorDataType]
    ) -> bool:
        """
        Given the previous and new measurement, it runs the comparison logic
        against the configured tolerance values and thresholds, and it
        processes the required sensor data change and above/below threshold
        events.

        :return: ``True`` if some events were processed for the new data,
            ``False`` otherwise.
        """
        # If the new data is missing or there are no changes, there are no
        # events to process
        if new_data is None or not self._has_changes(old_data, new_data):
            return False

        events = [
            SensorDataChangeEvent(
                data=new_data,  # type: ignore
                source=get_plugin_name_by_class(self.__class__),
            ),
            *self._process_threshold_events(old_data, new_data),
        ]

        for event in events:
            get_bus().post(event)

        self.publish_entities(new_data)
        return True

    def _update_last_measurement(self, new_data: SensorDataType):
        """
        Update the ``_last_measurement`` attribute with the newly acquired data.
        """
        # If there is no last measurement, or either the new or old
        # measurements are not dictionaries, then overwrite the previous data
        # with the new data
        if not (
            isinstance(self._last_measurement, dict) and isinstance(new_data, dict)
        ):
            self._last_measurement = new_data

        # Otherwise, merge the old data with the new
        self._last_measurement.update(new_data)  # type: ignore

    def _filter_enabled_sensors(self, data: SensorDataType) -> SensorDataType:
        """
        If ``data`` is a sensor mapping, and ``enabled_sensors`` is set, filter
        out only the requested keys.
        """
        if not (isinstance(data, dict) and self._enabled_sensors):
            return data

        return {k: v for k, v in data.items() if k in self._enabled_sensors}

    @abstractmethod
    def transform_entities(self, entities: SensorDataType) -> Collection[Entity]:
        raise NotImplementedError()

    def publish_entities(
        self, entities: SensorDataType, *args, **kwargs
    ) -> Collection[Entity]:
        # Skip empty data
        if (
            entities is None
            or (isinstance(entities, dict) and not entities)
            or (isinstance(entities, (list, tuple)) and not entities)
        ):
            return []

        return super().publish_entities(entities, *args, **kwargs)  # type: ignore

    @abstractmethod
    @action
    def get_measurement(self, *args, **kwargs) -> SensorDataType:
        """
        Implemented by the subclasses.

        :returns: Either a raw scalar::

            output = 273.16

        or a name-value dictionary with the values that have been read::

            output = {
                "temperature": 21.5,
                "humidity": 41.0
            }

        or a list of values::

            output = [
                0.01,
                0.34,
                0.53,
                ...
            ]

        """
        raise NotImplementedError()

    @action
    def get_data(self, *args, **kwargs):
        """
        (Deprecated) alias for :meth:`.get_measurement`
        """
        return self.get_measurement(*args, **kwargs)

    @action
    def status(self, *_, **__) -> Optional[SensorDataType]:
        """
        Returns the latest read values and publishes the
        :class:`platypush.message.event.entities.EntityUpdateEvent` events if
        required.
        """
        if self._last_measurement is not None:
            self.publish_entities(self._last_measurement)
        return self._last_measurement

    def main(self):
        sleep_retry_secs = 1  # Exponential back-off

        while not self.should_stop():
            try:
                new_data: SensorDataType = self.get_measurement().output  # type: ignore
                # Reset the exponential back-off retry counter in case of success
                sleep_retry_secs = 1
            except Exception as e:
                self.logger.warning(
                    'Could not update the status: %s. Next retry in %d seconds',
                    e,
                    sleep_retry_secs,
                )
                self.logger.exception(e)
                self.wait_stop(sleep_retry_secs)
                sleep_retry_secs = min(
                    sleep_retry_secs * 2,
                    self._max_retry_secs,
                )
                continue

            new_data = self._filter_enabled_sensors(new_data)
            is_event_processed = self._process_sensor_events(
                self._last_measurement, new_data
            )
            if is_event_processed:
                self._update_last_measurement(new_data)
            self.wait_stop(self.poll_interval)


# vim:sw=4:ts=4:et:



from typing import Union, Type, Callable

from .._parser import DocstringParser


class Constructor(DocstringParser):
    """
    Represents an integration constructor.
    """

    @classmethod
    def parse(cls, obj: Union[Type, Callable]) -> "Constructor":
        """
        Parse the parameters of a class constructor or action method.

        :param obj: Base type of the object.
        :return: The parsed parameters.
        """
        init = getattr(obj, "__init__", None)
        if init and callable(init):
            return super().parse(init)

        return super().parse(obj)



from typing import Iterable, Optional

from platypush.entities.bluetooth import BluetoothDevice
from platypush.message.event import Event


class BluetoothEvent(Event):
    """
    Base class for Bluetooth events.
    """


class BluetoothScanPausedEvent(BluetoothEvent):
    """
    Event triggered when the Bluetooth scan is paused.
    """

    def __init__(self, *args, duration: Optional[float] = None, **kwargs):
        super().__init__(*args, duration=duration, **kwargs)


class BluetoothScanResumedEvent(BluetoothEvent):
    """
    Event triggered when the Bluetooth scan is resumed.
    """

    def __init__(self, *args, duration: Optional[float] = None, **kwargs):
        super().__init__(*args, duration=duration, **kwargs)


class BluetoothDeviceEvent(BluetoothEvent):
    """
    Base class for Bluetooth device events.
    """

    def __init__(
        self,
        *args,
        address: str,
        connected: bool,
        name: Optional[str] = None,
        rssi: Optional[int] = None,
        tx_power: Optional[int] = None,
        manufacturer: Optional[str] = None,
        services: Optional[Iterable[dict]] = None,
        **kwargs
    ):
        """
        :param address: The Bluetooth address of the device.
        :param connected: Whether the device is connected.
        :param name: The name of the device.
        :param rssi: Received Signal Strength Indicator.
        :param tx_power: Transmission power.
        :param manufacturers: The manufacturers published by the device, as a
            ``manufacturer_id -> registered_name`` map.
        :param services: The services published by the device.
        """
        super().__init__(
            *args,
            address=address,
            name=name,
            connected=connected,
            rssi=rssi,
            tx_power=tx_power,
            manufacturer=manufacturer,
            services=services,
            **kwargs
        )

    @classmethod
    def from_device(cls, device: BluetoothDevice, **kwargs) -> "BluetoothDeviceEvent":
        """
        Initialize a Bluetooth event from the parameters of a device.

        :param device: Bluetooth device.
        """
        return cls(
            address=device.address,
            name=device.name,
            connected=device.connected,
            rssi=device.rssi,
            tx_power=device.tx_power,
            manufacturer=device.manufacturer,
            services=[srv.to_dict() for srv in device.services],
            **kwargs
        )


class BluetoothDeviceFoundEvent(BluetoothDeviceEvent):
    """
    Event triggered when a Bluetooth device is discovered during a scan.
    """


class BluetoothDeviceLostEvent(BluetoothDeviceEvent):
    """
    Event triggered when a previously discovered Bluetooth device is lost.
    """


class BluetoothDeviceConnectedEvent(BluetoothDeviceEvent):
    """
    Event triggered when a Bluetooth device is connected.
    """


class BluetoothDeviceDisconnectedEvent(BluetoothDeviceEvent):
    """
    Event triggered when a Bluetooth device is disconnected.
    """


class BluetoothDeviceSignalUpdateEvent(BluetoothDeviceEvent):
    """
    Event triggered when the RSSI/TX power of a Bluetooth device is updated.
    """


class BluetoothConnectionFailedEvent(BluetoothDeviceEvent):
    """
    Event triggered when a Bluetooth connection fails.
    """


class BluetoothFileEvent(BluetoothDeviceEvent):
    """
    Base class for Bluetooth file events.
    """

    def __init__(self, *args, file: str, **kwargs):
        super().__init__(*args, file=file, **kwargs)


class BluetoothFileTransferStartedEvent(BluetoothFileEvent):
    """
    Event triggered when a file transfer is initiated.
    """


class BluetoothFileTransferCancelledEvent(BluetoothFileEvent):
    """
    Event triggered when a file transfer is cancelled.
    """


class BluetoothFileReceivedEvent(BluetoothFileEvent):
    """
    Event triggered when a file download is completed.
    """


class BluetoothFileSentEvent(BluetoothFileEvent):
    """
    Event triggered when a file upload is completed.
    """


class BluetoothFilePutRequestEvent(BluetoothFileEvent):
    """
    Event triggered when a file put request is received.
    """


class BluetoothFileGetRequestEvent(BluetoothFileEvent):
    """
    Event triggered when a file get request is received.
    """


# vim:sw=4:ts=4:et:



import os
import importlib
import inspect
from typing import List, Type

import pkgutil

from ..ws import WSRoute, logger


def get_ws_routes() -> List[Type[WSRoute]]:
    """
    Scans for websocket route objects.
    """
    from platypush.backend.http import HttpBackend

    base_pkg = '.'.join([HttpBackend.__module__, 'app', 'ws'])
    base_dir = os.path.join(os.path.dirname(inspect.getfile(HttpBackend)), 'app', 'ws')
    routes = []

    for _, mod_name, _ in pkgutil.walk_packages([base_dir], prefix=base_pkg + '.'):
        try:
            module = importlib.import_module(mod_name)
        except Exception as e:
            logger.warning('Could not import module %s', mod_name)
            logger.exception(e)
            continue

        for _, obj in inspect.getmembers(module):
            if (
                inspect.isclass(obj)
                and not inspect.isabstract(obj)
                and issubclass(obj, WSRoute)
            ):
                routes.append(obj)

    return routes
